"""Unit tests for supports."""

from osmg.analysis.supports import ElasticSupport, FixedSupport


def test_fixed_support_initialization() -> None:
    """Test initialization of FixedSupport.

    Ensures that FixedSupport is correctly initialized with the
    specified degrees of freedom restraints.
    """
    support = FixedSupport((True, False, True))
    assert isinstance(support, FixedSupport)
    assert support == FixedSupport((True, False, True))


def test_elastic_support_initialization() -> None:
    """Test initialization of ElasticSupport.

    Ensures that ElasticSupport is correctly initialized with the
    specified degrees of freedom restraints.
    """
    support = ElasticSupport((10.0, 5.0, 0.0))
    assert isinstance(support, ElasticSupport)
    assert support == ElasticSupport((10.0, 5.0, 0.0))



"""objects that create materials."""

from dataclasses import dataclass

import numpy as np

from osmg.core.model import Model
from osmg.model_objects.section import ElasticSection, FiberSection
from osmg.model_objects.uniaxial_material import (
    Elastic,
    Fatigue,
    Hysteretic,
    IMKBilin,
    MaxStrainRange,
    Pinching4,
    Steel02,
    UniaxialMaterial,
)


@dataclass(repr=False)
class MaterialCreator:
    """Base class for uniaxial material creators."""

    model: Model

    def generate(self) -> UniaxialMaterial:  # noqa: D102, PLR6301
        msg = 'Subclasses must implement this method.'
        raise NotImplementedError(msg)


@dataclass(repr=False)
class ElasticMaterialCreator(MaterialCreator):
    """Generates an Elastic uniaxial material."""

    stiffness: float

    def generate(self) -> Elastic:
        """
        Generate a material.

        Returns:
          The material.
        """
        return Elastic(
            self.model.uid_generator,
            'Elastic',
            self.stiffness,
        )


@dataclass(repr=False)
class SteelHSSRectBraceMaxStrainRangeMaterialCreator(MaterialCreator):
    """
    Max strain range material.

    Sen, A. D., Roeder, C. W., Lehman, D. E., & Berman,
    J. W. (2019). Nonlinear modeling of concentrically braced
    frames. Journal of Constructional Steel Research, 157,
    103-120.

    """

    section: FiberSection
    e_mod: float
    g_mod: float
    f_y: float
    brace_length: float
    node_i_uid: int
    node_j_uid: int

    def generate(self) -> MaxStrainRange:
        """
        Generate a material.

        Returns:
          The material.
        """
        param_b = 0.001
        param_r0 = 15
        param_c_r1 = 0.925
        param_c_r2 = 0.15

        assert self.section.properties is not None
        sec_b = self.section.properties.B
        sec_t = self.section.properties.tdes
        var_lc = self.brace_length
        sec_r = min(self.section.properties.rx, self.section.properties.ry)
        mat_e = self.e_mod
        mat_g = self.g_mod
        mat_fy = self.f_y
        var_msr = (
            (0.554)
            * (sec_b / sec_t) ** (-0.75)
            * (var_lc / sec_r) ** (-0.47)
            * (mat_e / mat_fy) ** (0.21)
        )

        steel02_mat = Steel02(
            self.model.uid_generator,
            'auto_steel02_brace_mat',
            mat_fy,
            mat_e,
            mat_g,
            param_b,
            param_r0,
            param_c_r1,
            param_c_r2,
        )

        return MaxStrainRange(
            self.model.uid_generator,
            'auto_maxstrainrange_brace_mat',
            steel02_mat,
            var_msr,
            tangent_ratio=1.0e-5,
            def_coeff=0.068,
            node_tags=(self.node_i_uid, self.node_j_uid),
        )


@dataclass(repr=False)
class SteelHSSCircBraceFatigueMaterialCreator(MaterialCreator):
    """
    Circular HSS brace fatigue material.

    Karamanaci and Lignos (2014). Computational Approach for
    Collapse Assessment of Concentrically Braced Frames in Seismic
    Regions. Journal of Structural Engineering

    Returns:
      The defined material.
    """

    section: FiberSection
    e_mod: float
    g_mod: float
    f_y: float
    brace_length: float

    def generate(self) -> Fatigue:
        """
        Generate a material.

        Returns:
          The material.
        """
        param_b = 0.005
        param_r0 = 24.00
        param_c_r1 = 0.925
        param_c_r2 = 0.25

        assert self.section.properties is not None
        sec_d = self.section.properties.OD
        sec_t = self.section.properties.tdes
        var_lc = self.brace_length
        sec_r = min(self.section.properties.rx, self.section.properties.ry)
        mat_e = self.e_mod
        mat_g = self.g_mod
        mat_fy = self.f_y
        var_e0 = (
            (0.748)
            * (var_lc / sec_r) ** (-0.399)
            * (sec_d / sec_t) ** (-0.628)
            * (mat_e / mat_fy) ** (0.201)
        )
        var_m = -0.300

        steel02_mat = Steel02(
            self.model.uid_generator,
            'auto_steel02_brace_mat',
            mat_fy,
            mat_e,
            mat_g,
            param_b,
            param_r0,
            param_c_r1,
            param_c_r2,
            a1=0.2,
            a2=1.0,
            a3=0.2,
            a4=1.0,
        )

        return Fatigue(
            self.model.uid_generator,
            'auto_fatigue_brace_mat',
            steel02_mat,
            var_e0,
            var_m,
        )


@dataclass(repr=False)
class SteelWIMKMaterialCreator(MaterialCreator):
    """
    Steel W-shape IMK material.

    Assumes force unit is lb and length unit is inch.

    Lignos, D. G., & Krawinkler, H. (2011). Deterioration modeling of
    steel components in support of collapse prediction of steel moment
    frames under earthquake loading. Journal of Structural
    Engineering-Reston, 137(11), 1291.

    Elkady, A., & Lignos, D. G. (2014). Modeling of the composite
    action in fully restrained beam-to-column connections:
    implications in the seismic design and collapse capacity of steel
    special moment frames. Earthquake Engineering & Structural
    Dynamics, 43(13), 1935-1954.

    """

    section: ElasticSection
    f_y: float
    element_length: float
    lboverl: float
    loverh: float
    rbs_factor: float
    consider_composite: bool
    axial_load_ratio: float
    direction: str = 'strong'
    moment_modifier: float = 1.00
    n_parameter: float = 0.00

    def generate(self) -> IMKBilin:  # noqa: C901
        """
        Generate a material.

        Returns:
          The material.
        """
        # gather necessary data and check interpolation range
        assert self.section.name[0] == 'W', 'Error: Only W sections can be used.'
        assert isinstance(self.section, ElasticSection)
        assert self.section.properties
        # Yield stress
        mat_fy = self.f_y
        # Moment of inertia - strong axis - original section
        if self.direction == 'strong':
            sec_i = self.section.properties.Ix
        else:
            sec_i = self.section.properties.Iy
        # Section depth
        sec_d = self.section.properties.d
        # Flange width
        sec_bf = self.section.properties.bf
        # Flange and web thicknesses
        sec_tf = self.section.properties.tf
        sec_tw = self.section.properties.tw
        # Plastic modulus (unreduced)
        if self.direction == 'strong':
            sec_z = self.section.properties.Zx
        else:
            sec_z = self.section.properties.Zy
        # Radius of gyration
        sec_ry = self.section.properties.ry
        # Clear length
        elm_h = self.element_length
        # Shear span
        elm_l = self.loverh * elm_h
        elm_lb = self.lboverl * elm_l
        lbry = elm_lb / sec_ry

        # consider cases

        # checks ~ acceptable range
        if not 20.00 < sec_d / sec_tw < 55.00:  # noqa: PLR2004
            print(
                f'Warning: sec_d/sec_tw={sec_d / sec_tw:.2f}'
                ' outside regression range'
            )
            print('20.00 < sec_d/sec_tw < 55.00')
            print(self.section.name, '\n')
        if not 20.00 < lbry < 80.00:  # noqa: PLR2004
            print(f'Warning: Lb/ry={lbry:.2f} outside regression range')
            print('20.00 < lbry < 80.00')
            print(self.section.name, '\n')
        if not 4.00 < (sec_bf / (2.0 * sec_tf)) < 8.00:  # noqa: PLR2004
            print(
                f'Warning: bf/(2 tf)={sec_bf / (2. * sec_tf):.2f}'
                ' outside regression range'
            )
            print('4.00 < (sec_bf/(2.*sec_tf)) < 8.00')
            print(self.section.name, '\n')
        if not 2.5 < elm_l / sec_d < 7.0:  # noqa: PLR2004
            print(f'Warning: L/d={elm_l / sec_d:.2f}  outside regression range')
            print('2.5 < elm_l/sec_d < 7.0')
            print(self.section.name, '\n')
        if not 4.00 < sec_d < 36.00:  # noqa: PLR2004
            print(f'Warning: Section d={sec_d:.2f} outside regression range')
            print('4.00 < sec_d < 36.00')
            print(self.section.name, '\n')
        if not 35.00 < mat_fy < 65.00:  # noqa: PLR2004
            print(f'Warning: Fy={mat_fy:.2f} outside regression range')
            print('35.00 < mat_fy < 65.00')
            print(self.section.name, '\n')
        if self.rbs_factor:
            # RBS case
            assert self.direction == 'strong'
            assert self.rbs_factor <= 1.00, 'rbs_factor must be <= 1.00'
            # calculate parameters
            theta_p = (
                0.19
                * (sec_d / sec_tw) ** (-0.314)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.10)
                * lbry ** (-0.185)
                * (elm_l / sec_d) ** 0.113
                * (25.4 * sec_d / 533.0) ** (-0.76)
                * (6.895 * mat_fy / 355.0) ** (-0.07)
            )
            theta_pc = (
                9.52
                * (sec_d / sec_tw) ** (-0.513)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.863)
                * lbry ** (-0.108)
                * (6.895 * mat_fy / 355.0) ** (-0.36)
            )
            lambda_var = (
                585.0
                * (sec_d / sec_tw) ** (-1.14)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.632)
                * lbry ** (-0.205)
                * (6.895 * mat_fy / 355.0) ** (-0.391)
            )
            rbs_c = sec_bf * (1.0 - self.rbs_factor) / 2.0
            z_rbs = sec_z - 2.0 * rbs_c * sec_tf * (sec_d - sec_tf)
            sec_m = 1.06 * z_rbs * mat_fy
            mcmy_plus = 1.10
            mcmy_minus = 1.10

        # Other-than-RBS case
        elif self.axial_load_ratio:
            # column case
            theta_p = (
                294.00
                * (sec_d / sec_tw) ** (-1.70)
                * lbry ** (-0.70)
                * (1.00 - self.axial_load_ratio) ** (1.60)
            )
            theta_pc = (
                90.00
                * (sec_d / sec_tw) ** (-0.80)
                * lbry ** (-0.80)
                * (1.00 - self.axial_load_ratio) ** (2.50)
            )
            theta_p = min(theta_p, 0.20)
            theta_pc = min(theta_pc, 0.30)
            if self.axial_load_ratio <= 0.35:  # noqa: PLR2004
                lambda_var = (
                    25500.00
                    * (sec_d / sec_tw) ** (-2.14)
                    * lbry ** (-0.53)
                    * (1.00 - self.axial_load_ratio) ** (4.29)
                )
            else:
                lambda_var = (
                    268000.00
                    * (sec_d / sec_tw) ** (-2.30)
                    * lbry ** (-1.30)
                    * (1.00 - self.axial_load_ratio) ** (1.19)
                )
            if self.axial_load_ratio <= 0.20:  # noqa: PLR2004
                sec_m = (
                    1.15
                    / 1.10
                    * (sec_z * mat_fy)
                    * (1.00 - self.axial_load_ratio / 2.00)
                )
            else:
                sec_m = (
                    1.15
                    / 1.10
                    * (sec_z * mat_fy)
                    * 9.0
                    / 8.0
                    * (1.00 - self.axial_load_ratio)
                )
            mcmy = (
                12.5
                * (sec_d / sec_tw) ** (-0.20)
                * lbry ** (-0.40)
                * (1.00 - self.axial_load_ratio) ** (0.40)
            )
            mcmy = min(mcmy, 1.00)
            mcmy = max(mcmy, 1.30)
            mcmy_plus = mcmy
            mcmy_minus = mcmy

        else:
            # non-RBS beam case
            theta_p = (
                0.0865
                * (sec_d / sec_tw) ** (-0.365)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.14)
                * (elm_l / sec_d) ** 0.34
                * (25.4 * sec_d / 533.0) ** (-0.721)
                * (6.895 * mat_fy / 355.0) ** (-0.23)
            )
            theta_pc = (
                5.63
                * (sec_d / sec_tw) ** (-0.565)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.800)
                * (25.4 * sec_d / 533.0) ** (-0.28)
                * (6.895 * mat_fy / 355.0) ** (-0.43)
            )
            lambda_var = (
                495.0
                * (sec_d / sec_tw) ** (-1.34)
                * (sec_bf / (2.0 * sec_tf)) ** (-0.595)
                * (6.895 * mat_fy / 355.0) ** (-0.36)
            )
            sec_m = 1.17 * sec_z * mat_fy
            mcmy_plus = 1.10
            mcmy_minus = 1.10

        theta_u = 0.20
        residual_plus = 0.40
        residual_minus = 0.40
        theta_p_plus = theta_p
        theta_p_minus = theta_p
        theta_pc_plus = theta_pc
        theta_pc_minus = theta_pc
        d_plus = 1.00
        d_minus = 1.00
        m_plus = sec_m
        m_minus = -sec_m

        if self.consider_composite:
            # Elkady, A., & Lignos, D. G. (2014). Modeling of the
            # composite action in fully restrained beam-to-column
            # connections: implications in the seismic design and
            # collapse capacity of steel special moment
            # frames. Earthquake Engineering & Structural Dynamics,
            # 43(13), 1935-1954.  Table II

            assert (
                self.axial_load_ratio == 0.00
            ), "Can't consider composite action for columns"
            assert (
                self.direction == 'strong'
            ), 'Composite action affects the behavior in strong-axis bending'

            theta_p_plus *= 1.80
            theta_p_minus *= 0.95
            theta_pc_plus *= 1.35
            theta_pc_minus *= 0.95
            d_plus = 1.15
            d_minus = 1.00
            mcmy_plus = 1.30
            mcmy_minus = 1.05
            m_plus *= 1.35
            m_minus *= 1.25
            residual_plus = 0.30
            residual_minus = 0.20

        # adjust parameters to account for the presence of the elastic element
        stiffness_init = 6.00 * self.section.e_mod * sec_i / elm_h
        stiffness = (self.n_parameter + 1.00) * stiffness_init
        theta_y = sec_m / stiffness_init
        theta_p_plus -= (mcmy_plus - 1.0) * (sec_m / stiffness)
        theta_p_minus -= (mcmy_minus - 1.0) * (sec_m / stiffness)
        theta_pc_plus += theta_y + (mcmy_plus - 1.0) * (sec_m / stiffness)
        theta_pc_plus += theta_y + (mcmy_minus - 1.0) * (sec_m / stiffness)
        beta_plus = (mcmy_plus - 1.0) * m_plus / theta_p_plus / stiffness
        beta_minus = -(mcmy_minus - 1.0) * m_minus / theta_p_minus / stiffness

        # new model
        return IMKBilin(
            self.model.uid_generator,
            'auto_IMK',
            stiffness * self.moment_modifier,
            theta_p_plus,
            theta_pc_plus,
            theta_u,
            m_plus * self.moment_modifier,
            (1.0 + beta_plus),
            residual_plus,
            theta_p_minus,
            theta_pc_minus,
            theta_u,
            -m_minus * self.moment_modifier,
            (1.0 + beta_minus),
            residual_minus,
            lambda_var,
            lambda_var,
            lambda_var,
            1.00,
            1.00,
            1.00,
            d_plus,
            d_minus,
        )


@dataclass(repr=False)
class SteelGravityShearTabCreator(MaterialCreator):
    """
    Gravity shear tab connection hinge.

    Assumes force unit is lb and length unit is inch.

    Elkady, A., & Lignos, D. G. (2015). Effect of gravity framing on
    the overstrength and collapse capacity of steel frame buildings
    with perimeter special moment frames. Earthquake Engineering &
    Structural Dynamics, 44(8), 1289-1307.
    """

    section: ElasticSection
    f_y: float
    consider_composite: bool
    moment_modifier: float = 1.00

    def generate(self) -> Pinching4:
        """
        Generate a material.

        Returns:
          The material.
        """
        assert self.section.name[0] == 'W', 'Error: Only W sections can be used.'
        assert isinstance(self.section, ElasticSection)
        assert self.section.properties

        # Yield stress
        assert isinstance(self.moment_modifier, float)
        mat_fy = self.f_y
        # Plastic modulus (unreduced)
        sec_zx = self.section.properties.Zx
        # Plastic moment of the section
        sec_mp = sec_zx * mat_fy * self.moment_modifier

        if not self.consider_composite:
            m_max_pos = 0.121 * sec_mp
            m_max_neg = 0.121 * sec_mp
            m1_p = +0.521 * m_max_pos
            m1_n = -0.521 * m_max_neg
            m2_p = +0.967 * m_max_pos
            m2_n = -0.967 * m_max_neg
            m3_p = +1.000 * m_max_pos
            m3_n = -1.000 * m_max_pos
            m4_p = +0.901 * m_max_pos
            m4_n = -0.901 * m_max_neg
            th_1_p = 0.0045
            th_1_n = -0.0045
            th_2_p = 0.0465
            th_2_n = -0.0465
            th_3_p = 0.0750
            th_3_n = -0.0750
            th_4_p = 0.1000
            th_4_n = -0.1000
            rdispp = 0.57
            rdispn = 0.57
            rforcep = 0.40
            rforcen = 0.40
            uforcep = 0.05
            uforcen = 0.05
            gklim = 0.2
            gdlim = 0.1
            gflim = 0.0
            g_e = 10
            dmgtype = 'energy'
        else:
            m_max_pos = 0.35 * sec_mp
            m_max_neg = 0.64 * 0.35 * sec_mp
            m1_p = +0.250 * m_max_pos
            m1_n = -0.250 * m_max_neg
            m2_p = +1.000 * m_max_pos
            m2_n = -1.000 * m_max_neg
            m3_p = +1.001 * m_max_pos
            m3_n = -1.001 * m_max_pos
            m4_p = +0.530 * m_max_pos
            m4_n = -0.540 * m_max_neg
            th_1_p = 0.0042
            th_1_n = -0.0042
            th_2_p = 0.0200
            th_2_n = -0.0110
            th_3_p = 0.0390
            th_3_n = -0.0300
            th_4_p = 0.0400
            th_4_n = -0.0550
            rdispp = 0.40
            rdispn = 0.50
            rforcep = 0.13
            rforcen = 0.53
            uforcep = 0.01
            uforcen = 0.05
            gklim = 0.30
            gdlim = 0.05
            gflim = 0.05
            g_e = 10
            dmgtype = 'energy'

        return Pinching4(
            self.model.uid_generator,
            'auto_gravity_shear_tab',
            m1_p,
            th_1_p,
            m2_p,
            th_2_p,
            m3_p,
            th_3_p,
            m4_p,
            th_4_p,
            m1_n,
            th_1_n,
            m2_n,
            th_2_n,
            m3_n,
            th_3_n,
            m4_n,
            th_4_n,
            rdispp,
            rforcep,
            uforcep,
            rdispn,
            rforcen,
            uforcen,
            0.00,
            0.00,
            0.00,
            0.00,
            gklim,
            0.00,
            0.00,
            0.00,
            0.00,
            gdlim,
            0.00,
            0.00,
            0.00,
            0.00,
            gflim,
            g_e,
            dmgtype,
        )


@dataclass(repr=False)
class SteelWColumnPanelZoneCreator(MaterialCreator):
    """
    Panel zone hinge (parallelogram model).

    Assumes force unit is lb and length unit is inch.

    Gupta, A., & Krawinkler, H. (1999). Seismic demands for the
    performance evaluation of steel moment resisting frame
    structures. Rep. No. 132.
    """

    section: ElasticSection
    f_y: float
    g_mod: float
    pz_length: float
    pz_doubler_plate_thickness: float
    pz_hardening: float
    moment_modifier: float = 1.00

    def generate(self) -> Hysteretic:
        """
        Generate a material.

        Returns:
          The material.
        """
        assert self.section.name[0] == 'W', 'Error: Only W sections can be used.'
        assert isinstance(self.section, ElasticSection)
        assert self.section.properties
        f_y = self.f_y
        hardening = self.pz_hardening
        d_c = self.section.properties.d
        bfc = self.section.properties.bf
        t_p = self.section.properties.tw + self.pz_doubler_plate_thickness
        t_f = self.section.properties.tf
        v_y = 0.55 * f_y * d_c * t_p
        g_mod = self.g_mod
        k_e = 0.95 * g_mod * t_p * d_c
        k_p = 0.95 * g_mod * bfc * t_f**2 / self.pz_length
        gamma_1 = v_y / k_e
        gamma_2 = 4.0 * gamma_1
        gamma_3 = 100.0 * gamma_1
        m1y = gamma_1 * k_e * self.pz_length * self.moment_modifier
        m2y = m1y + k_p * self.pz_length * (gamma_2 - gamma_1) * self.moment_modifier
        m3y = (
            m2y
            + (hardening * k_e * self.pz_length)
            * (gamma_3 - gamma_2)
            * self.moment_modifier
        )

        return Hysteretic(
            self.model.uid_generator,
            'auto_steel_W_PZ',
            (m1y, gamma_1),
            (m2y, gamma_2),
            (m3y, gamma_3),
            (-m1y, -gamma_1),
            (-m2y, -gamma_2),
            (-m3y, -gamma_3),
            0.25,
            0.75,
            0.00,
            0.00,
            0.00,
        )


@dataclass(repr=False)
class SteelWColumnPanelZoneUpdatedCreator(MaterialCreator):
    """
    Define updated panel zone hinge (parallelogram model).

    Assumes force unit is lb and length unit is inch.

    Skiadopoulos, A., Elkady, A. and D. G. Lignos (2020). "Proposed
    Panel Zone Model for Seismic Design of Steel Moment-Resisting
    Frames." ASCE Journal of Structural Engineering. DOI:
    10.1061/(ASCE)ST.1943-541X.0002935.

    """

    section: ElasticSection
    e_mod: float
    g_mod: float
    f_y: float
    pz_length: float
    pz_doubler_plate_thickness: float
    axial_load_ratio: float
    slab_depth: float
    location: str
    consider_composite: bool
    moment_modifier: float = 1.00

    def generate(self) -> Hysteretic:  # noqa: PLR0914
        """
        Generate a material.

        Returns:
          The material.

        Raises:
          ValueError: If the specified `location` is invalid.
        """
        assert self.section.name[0] == 'W', 'Error: Only W sections can be used.'
        assert isinstance(self.section, ElasticSection)
        assert self.section.properties
        f_y = self.f_y
        e_mod = self.e_mod
        g_mod = self.g_mod
        tw_Col = self.section.properties.tw  # noqa: N806
        tdp = self.pz_doubler_plate_thickness
        d_Col = self.section.properties.d  # noqa: N806
        d_Beam = self.pz_length  # noqa: N806
        tf_Col = self.section.properties.tf  # noqa: N806
        bf_Col = self.section.properties.bf  # noqa: N806
        Ix_Col = self.section.properties.Ix  # noqa: N806
        ts = self.slab_depth
        n = self.axial_load_ratio
        trib = self.slab_depth

        tpz = tw_Col + tdp  # total PZ thickness
        # effective depth in positive moment
        d_BeamP = d_Beam  # noqa: N806
        if self.consider_composite:
            d_BeamP = d_Beam + trib + 0.50 * ts  # noqa: N806
        # effective depth in negative moment
        d_BeamN = d_Beam  # noqa: N806

        # Stiffness Calculation
        Ks = tpz * (d_Col - tf_Col) * g_mod  # noqa: N806
        Kb = (  # noqa: N806
            12.0
            * e_mod
            * (Ix_Col + tdp * ((d_Col - 2.0 * tf_Col) ** 3) / 12.00)
            / (d_Beam**3)
            * d_Beam
        )
        Ke = Ks * Kb / (Ks + Kb)  # noqa: N806

        # flange stiffness: shear contribution
        Ksf = 2.0 * (bf_Col * tf_Col) * g_mod  # noqa: N806
        # flange stiffness: bending contribution
        Kbf = 2.0 * 12.0 * e_mod * bf_Col * (tf_Col**3) / 12.0 / (d_Beam**3) * d_Beam  # noqa: N806
        # flange stiffness: total contribution
        Kef = (Ksf * Kbf) / (Ksf + Kbf)  # noqa: N806

        ay = (0.58 * Kef / Ke + 0.88) / (1.0 - Kef / Ke)

        aw_eff_4gamma = 1.10
        aw_eff_6gamma = 1.15

        af_eff_4gamma = 0.93 * Kef / Ke + 0.015
        af_eff_6gamma = 1.05 * Kef / Ke + 0.020
        # reduction factor accounting for axial load
        r = np.sqrt(1.0 - (n**2))

        Vy = r * 0.577 * f_y * ay * (d_Col - tf_Col) * tpz  # noqa: N806
        # Plastic Shear Force at 4 gammaY
        Vp_4gamma = (  # noqa: N806
            r
            * 0.577
            * f_y
            * (
                aw_eff_4gamma * (d_Col - tf_Col) * tpz
                + af_eff_4gamma * (bf_Col - tw_Col) * 2 * tf_Col
            )
        )
        # Plastic Shear Force at 6 gammaY
        Vp_6gamma = (  # noqa: N806
            r
            * 0.577
            * f_y
            * (
                aw_eff_6gamma * (d_Col - tf_Col) * tpz
                + af_eff_6gamma * (bf_Col - tw_Col) * 2 * tf_Col
            )
        )

        gamma_y = Vy / Ke
        gamma4_y = 4.0 * gamma_y

        My_P = Vy * d_BeamP  # noqa: N806
        Mp_4gamma_P = Vp_4gamma * d_BeamP  # noqa: N806
        Mp_6gamma_P = Vp_6gamma * d_BeamP  # noqa: N806

        My_N = Vy * d_BeamN  # noqa: N806
        Mp_4gamma_N = Vp_4gamma * d_BeamN  # noqa: N806
        Mp_6gamma_N = Vp_6gamma * d_BeamN  # noqa: N806

        Slope_4to6gamma_y_P = (Mp_6gamma_P - Mp_4gamma_P) / (2.0 * gamma_y)  # noqa: N806
        Slope_4to6gamma_y_N = (Mp_6gamma_N - Mp_4gamma_N) / (2.0 * gamma_y)  # noqa: N806

        # Defining the 3 Points used to construct the trilinear backbone curve
        gamma1 = gamma_y
        gamma2 = gamma4_y
        gamma3 = 100.0 * gamma_y

        M1_P = My_P  # noqa: N806
        M2_P = Mp_4gamma_P  # noqa: N806
        M3_P = Mp_4gamma_P + Slope_4to6gamma_y_P * (100 * gamma_y - gamma4_y)  # noqa: N806

        M1_N = My_N  # noqa: N806
        M2_N = Mp_4gamma_N  # noqa: N806
        M3_N = Mp_4gamma_N + Slope_4to6gamma_y_N * (100 * gamma_y - gamma4_y)  # noqa: N806

        # apply moment modifier
        M1_P *= self.moment_modifier  # noqa: N806
        M2_P *= self.moment_modifier  # noqa: N806
        M3_P *= self.moment_modifier  # noqa: N806
        M1_N *= self.moment_modifier  # noqa: N806
        M2_N *= self.moment_modifier  # noqa: N806
        M3_N *= self.moment_modifier  # noqa: N806

        if not self.consider_composite:
            args = (
                (M1_N, gamma1),
                (M2_N, gamma2),
                (M3_N, gamma3),
                (-M1_N, -gamma1),
                (-M2_N, -gamma2),
                (-M3_N, -gamma3),
                0.25,
                0.75,
                0.0,
                0.0,
                0.0,
            )
        elif self.location == 'interior':
            args = (
                (M1_P, gamma1),
                (M2_P, gamma2),
                (M3_P, gamma3),
                (-M1_P, -gamma1),
                (-M2_P, -gamma2),
                (-M3_P, -gamma3),
                0.25,
                0.75,
                0.0,
                0.0,
                0.0,
            )
        elif self.location == 'exterior_first':
            args = (
                (M1_N, gamma1),
                (M2_N, gamma2),
                (M3_N, gamma3),
                (-M1_P, -gamma1),
                (-M2_P, -gamma2),
                (-M3_P, -gamma3),
                0.25,
                0.75,
                0.0,
                0.0,
                0.0,
            )
        elif self.location == 'exterior_last':
            args = (
                (M1_P, gamma1),
                (M2_P, gamma2),
                (M3_P, gamma3),
                (-M1_N, -gamma1),
                (-M2_N, -gamma2),
                (-M3_N, -gamma3),
                0.25,
                0.75,
                0.0,
                0.0,
                0.0,
            )
        else:
            msg = f'Invalid Location: {self.location}'
            raise ValueError(msg)

        return Hysteretic(
            self.model.uid_generator,
            'auto_steel_W_pz_updated',
            *args,
        )


class SteelBraceGussetCreator(MaterialCreator):
    """
    Steel brace gusset plate hinge.

    Hsiao, P-C., Lehman, D.E., and Roeder, C.W., 2012, Improved
    analytical model for special concentrically braced frames, Journal
    of Constructional Steel Research, Vol. 73, pp 80-94.

    Arguments:
      model: Model object.
      d_brace: brace section height.
      l_c: brace-to-gusset connection length.
      t_p: gusset plate thickness.
      l_b: gusset plate average buckling length.
      **kwargs: Other keyword arguments.

    """

    model: Model
    e_mod: float
    g_mod: float
    f_y: float
    d_brace: float
    l_c: float
    t_p: float
    l_b: float

    def generate(self) -> Steel02:
        """
        Generate a material.

        Returns:
          The material.
        """
        var_w = self.d_brace + 2.00 * self.l_c * np.tan(30.00 / 180.00 * np.pi)
        var_i = var_w * self.t_p**3 / 12.00
        var_z = var_w * self.t_p**2 / 6.00
        f_y = self.f_y
        var_e = self.e_mod
        var_g = self.g_mod
        var_my = var_z * f_y
        var_k_rot = var_e * var_i / self.l_b
        var_b = 0.01
        return Steel02(
            self.model.uid_generator,
            'auto_steel_gusset',
            var_my,
            var_k_rot,
            var_g,
            var_b,
            20.00,
            0.925,
            0.15,
            0.0005,
            0.014,
            0.0005,
            0.01,
            0.00,
        )



"""
Braced frame design model.

Length units: in
Force units: kip
"""

import numpy as np

from osmg.analysis.common import UDL, PointLoad, PointMass
from osmg.analysis.load_case import LoadCaseRegistry
from osmg.analysis.supports import FixedSupport
from osmg.core.model import Model2D
from osmg.creators.component import BarGenerator, BeamColumnCreator
from osmg.creators.material import ElasticMaterialCreator
from osmg.creators.section import AISC_Database_Section_Creator
from osmg.graphics.plotly import (
    BasicForceConfiguration,
    DeformationConfiguration,
    Figure3D,
    Figure3DConfiguration,
)
from osmg.model_objects.node import Node
from osmg.model_objects.section import ElasticSection

# Instantiate model object
frame = Model2D(name='Gridline B', dimensionality='2D Frame')

# Add levels and grids
grids = frame.grid_system
grids.add_level('0', 0.00)
grids.add_level('1', 15.00 * 12.00)
grids.add_level('2', (15.00 + 13.00) * 12.00)
grids.add_level('3', (15.00 + 2.0 * 13.00) * 12.00)
grids.add_level('4', (15.00 + 3.0 * 13.00) * 12.00)
grids.add_grid('A', 0.00)
grids.add_grid('B', 25.00 * 12.00)
grids.add_grid('C', 25.00 * 2.0 * 12.00)
grids.add_grid('D', 25.00 * 3.0 * 12.00)
grids.add_grid('E', 25.00 * 4.0 * 12.00)
grids.add_grid('F', 25.00 * 5.0 * 12.00)
grids.add_grid('LC', 25.00 * 6.0 * 12.00)

# Add primary nodes
for level in ('0', '1', '2', '3', '4'):
    for grid in ('A', 'B', 'C', 'D', 'E', 'F', 'LC'):
        if (grid == 'A' and level == '4') or (grid == 'F' and level == '4'):
            continue
        frame.nodes.add(
            Node(
                uid_generator=frame.uid_generator,
                coordinates=(
                    grids.get_grid_location(grid),
                    grids.get_level(level).elevation(),
                ),
            ),
        )

# Example: find the node at 'Base'-'A' (not used)
found_node = frame.nodes.search_by_coordinates_or_raise(
    (
        grids.get_grid('A').data(),
        grids.get_level('0').elevation(),
    )
)

# Define a common section (not used)
simple_section = ElasticSection(
    frame.uid_generator,
    'Test Section',
    e_mod=1e3,
    area=1e3,
    i_y=1.00,
    i_x=1.00,
    g_mod=1.00,
    j_mod=1.00,
    sec_w=0.00,
)

# Define Steel Young's modulus and shear modulus.
e_modulus = 29000.00  # kip/in2
g_modulus = 11500.00  # kip/in2

# Define an AISC W section (length unit is `in`)
section_creator = AISC_Database_Section_Creator(frame.uid_generator)
column_section = section_creator.load_elastic_section(
    section_label='W14X120', e_modulus=e_modulus, g_modulus=g_modulus
)
beam_section = section_creator.load_elastic_section(
    section_label='W18X119', e_modulus=e_modulus, g_modulus=g_modulus
)


# Add columns
bcg = BeamColumnCreator(frame, 'elastic')
added_columns = []
for level in ('1', '2', '3', '4'):
    for grid in ('A', 'B', 'C', 'D', 'E', 'F'):
        if level == '4' and (grid in {'A', 'F'}):
            continue
        col = bcg.generate_plain_component_assembly(
            tags={'Column'},
            node_i=frame.nodes.search_by_coordinates_or_raise(
                (
                    grids.get_grid(grid).data(),
                    grids.get_level(level).elevation(),
                )
            ),
            node_j=frame.nodes.search_by_coordinates_or_raise(
                (
                    grids.get_grid(grid).data(),
                    grids.get_level(level).previous().elevation(),
                )
            ),
            n_sub=1,
            eo_i=np.array((0.00, 0.0)),
            eo_j=np.array((0.00, 0.0)),
            section=column_section,
            transf_type='Linear',
        )
        added_columns.append(col)

# Add beams
added_beams = []
for level in ('1', '2', '3', '4'):
    for grid in ('A', 'B', 'C', 'D', 'E'):
        if level == '4' and grid in {'A', 'E'}:
            continue
        beam = bcg.generate_plain_component_assembly(
            tags={'Beam'},
            node_i=frame.nodes.search_by_coordinates_or_raise(
                (
                    grids.get_grid(grid).data(),
                    grids.get_level(level).elevation(),
                )
            ),
            node_j=frame.nodes.search_by_coordinates_or_raise(
                (
                    grids.get_grid(grid).next().data(),
                    grids.get_level(level).elevation(),
                )
            ),
            n_sub=1,
            eo_i=np.array((0.00, 0.0)),
            eo_j=np.array((0.00, 0.0)),
            section=beam_section,
            transf_type='Linear',
        )
        added_beams.append(beam)

# # Add braces
brg = BarGenerator(frame)
for level_top, level_bottom, grid_top, grid_bottom in zip(
    ('1', '1', '1', '1', '2', '2', '3', '4'),
    ('0', '0', '0', '0', '1', '1', '2', '3'),
    ('C', 'C', 'E', 'E', 'D', 'D', 'C', 'D'),
    ('B', 'D', 'D', 'F', 'C', 'E', 'D', 'C'),
):
    brg.add(
        tags={'Brace'},
        node_i=frame.nodes.search_by_coordinates_or_raise(
            (
                grids.get_grid(grid_top).data(),
                grids.get_level(level_top).elevation(),
            )
        ),
        node_j=frame.nodes.search_by_coordinates_or_raise(
            (
                grids.get_grid(grid_bottom).data(),
                grids.get_level(level_bottom).elevation(),
            )
        ),
        eo_i=np.array((0.00, 0.00)),
        eo_j=np.array((0.00, 0.00)),
        transf_type='Linear',
        area=40.00,
        material=ElasticMaterialCreator(frame, stiffness=e_modulus).generate(),
        outside_shape=None,
        weight_per_length=0.00,
    )


# Add leaning column
leaning_column_area = 1e4  # in2
for level in ('1', '2', '3', '4'):
    brg.add(
        tags={'Truss'},
        node_i=frame.nodes.search_by_coordinates_or_raise(
            (
                grids.get_grid('LC').data(),
                grids.get_level(level).elevation(),
            )
        ),
        node_j=frame.nodes.search_by_coordinates_or_raise(
            (
                grids.get_grid('LC').data(),
                grids.get_level(level).previous().elevation(),
            )
        ),
        eo_i=np.array((0.00, 0.00)),
        eo_j=np.array((0.00, 0.00)),
        transf_type='Linear',
        area=leaning_column_area,
        material=ElasticMaterialCreator(frame, stiffness=e_modulus).generate(),
        outside_shape=None,
        weight_per_length=0.00,
    )


def show_model() -> None:
    """Show the model (only)."""
    fig = Figure3D(Figure3DConfiguration(ndm=2))
    fig.add_nodes(list(frame.nodes.values()), 'primary')
    fig.add_components(list(frame.components.values()))
    fig.show()


# Create a load case registry
load_case_registry = LoadCaseRegistry(frame)
# Define load cases
lc_modal = load_case_registry.modal['modal_1']
lc_dead = load_case_registry.dead['dead_1']
load_cases = (lc_modal, lc_dead)

# Create a load case and add fixed supports
fixed_support = FixedSupport((True, True, True))
for lc in load_cases:
    lc.add_supports_at_level(frame, fixed_support, '0')

# Add rigid diaphragm
for level in ('1', '2', '3', '4'):
    parent_node = frame.nodes.search_by_coordinates_or_raise(
        (
            grids.get_grid('LC').data(),
            grids.get_level(level).elevation(),
        )
    )
    for lc in load_cases:
        lc.define_rigid_diaphragm(frame, parent_node)


# # Example of how to retrieve a primary node:
# # Locate the nodes at 'A'-'Level 1' and 'B'-'Level 1'
# frame.nodes.search_by_coordinates_or_raise(
#     (grids.get_grid('A').data(), grids.get_level(').elevation())
# )

# `lc_dead`: Add UDLs to the beams
for beam in added_beams:
    lc_dead.load_registry.component_udl[beam.uid] = UDL((0.0, -1.67e-3))  # kip/in

# `lc_dead`: Add a concentrated point load at 'B'-'Level 4'
lc_dead.load_registry.nodal_loads[
    frame.nodes.search_by_coordinates_or_raise(
        (
            grids.get_grid('B').data(),
            grids.get_level('4').elevation(),
        )
    ).uid
] = PointLoad(
    (2000.0, 0.00, 0.00)  # lb
)

# `modal`: Add mass on the leaning column.
for level in ('1', '2', '3', '4'):
    lc_modal.mass_registry[
        frame.nodes.search_by_coordinates_or_raise(
            (
                grids.get_grid('LC').data(),
                grids.get_level(level).elevation(),
            )
        ).uid
    ] = PointMass((+16.0e3 / 386.22, 0.00, 0.00))
lc_modal.analysis.settings.num_modes = 2

# # Example: Add an extra recorder
# load_case_registry.dead['dead_1'].analysis.recorders['node_envelope'] = NodeRecorder(
#     uid_generator=frame.uid_generator,
#     recorder_type='EnvelopeNode',
#     nodes=(
#         frame.nodes.search_by_coordinates_or_raise(
#             (
#                 grids.get_grid('A').data(),
#                 grids.get_level('3').elevation(),
#             )
#         ).uid,
#     ),
#     dofs=(1, 2, 3),
#     response_type='disp',
#     file_name='envelope',
#     output_time=True,
#     number_of_significant_digits=6,
# )


# # Example: change num_steps
# load_case_registry.dead['dead_1'].analysis.settings.num_steps = 10

# Run analysis
load_case_registry.run()

# result_dir = load_case_registry.dead['dead_1'].analysis.settings.result_directory
# print(f'Result directory `dead_1`: {result_dir}')
# result_dir = load_case_registry.dead['dead_2'].analysis.settings.result_directory
# print(f'Result directory `dead_2`: {result_dir}')


# lc_modal.analysis.recorders['default_node'].get_data()
# lc_modal.analysis.recorders[
#     'default_basic_force'
# ].get_data()

# combinations happen here.
# combined_displacements = load_case_registry.combine_recorder('default_node')

# forces = (
#     load_case_registry.dead['dead_1']
#     .analysis.recorders['default_basic_force']
#     .get_data()
# )


# data = (
#     load_case_registry.dead['dead_1']
#     .analysis.recorders['default_basic_force']
#     .get_data()
# )


step = 0
deformation_configuration = DeformationConfiguration(
    reference_length=frame.reference_length(),
    ndf=3,
    ndm=2,
    data=lc_dead.analysis.recorders['default_node'].get_data(),
    step=step,
    # amplification_factor=None,  # Figure it out.
    amplification_factor=10.00 / 3.06618,  # Figure it out.
)
basic_force_configuration = BasicForceConfiguration(
    reference_length=frame.reference_length(),
    ndm=2,
    ndf=3,
    data=lc_dead.calculate_basic_forces(
        'default_basic_force',
        frame.components.get_line_element_lengths(),
        ndm=2,
        num_stations=12,
    ),
    step=step,
    force_to_length_factor=12 * 12 / 1883.89,
    moment_to_length_factor=12 * 12 / 10000.0,
)
fig = Figure3D(Figure3DConfiguration(ndm=2))
fig.add_nodes(list(frame.nodes.values()), 'primary', overlay=True)
fig.add_components(list(frame.components.values()), overlay=True)
# fig.add_nodes(list(frame.nodes.values()), 'primary')
# fig.add_components(list(frame.components.values()))
fig.add_nodes(list(frame.nodes.values()), 'primary', deformation_configuration)
fig.add_components(list(frame.components.values()), deformation_configuration)
fig.add_supports(frame.nodes, lc_modal.fixed_supports, symbol_size=12.00)
# fig.add_udl(
#     load_case_registry.dead['dead_2'].load_registry.component_udl,
#     frame.components,
#     force_to_length_factor=2.0,
#     offset=0.00,
# )
# fig.add_loads(
#     lc_modal.mass_registry,
#     frame.nodes,
#     force_to_length_factor=2.00,
#     offset=0.0,
#     head_length=24.0,
#     head_width=24.0,
#     base_width=5.0,
# )
fig.add_basic_forces(
    components=list(frame.components.values()),
    basic_force_configuration=basic_force_configuration,
)
fig.show()



"""objects that create component assemblies for a model."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

import numpy as np

from osmg.core.common import EPSILON, STIFF, STIFF_ROT
from osmg.core.osmg_collections import (
    BarAssembly,
    BeamColumnAssembly,
    ComponentAssembly,
)
from osmg.creators.material import ElasticMaterialCreator
from osmg.creators.zerolength import ZeroLengthCreator
from osmg.geometry.transformations import (
    local_axes_from_points_and_angle,
    transformation_matrix,
)
from osmg.model_objects.element import (
    Bar,
    DispBeamColumn,
    ElasticBeamColumn,
    GeomTransf,
    ModifiedStiffnessParameterConfig,
    TwoNodeLink,
)
from osmg.model_objects.node import Node
from osmg.model_objects.section import ElasticSection, FiberSection

if TYPE_CHECKING:
    from osmg.core.common import numpy_array
    from osmg.core.model import Model
    from osmg.creators.material import MaterialCreator
    from osmg.geometry.mesh import Mesh
    from osmg.model_objects.uniaxial_material import UniaxialMaterial


@dataclass(repr=False)
class BaseCreator:
    """Base class for component creators."""

    model: Model = field(repr=False)

    def define_zerolength_fixed_material_creators(
        self,
    ) -> dict[int, MaterialCreator]:
        """
        Define material creators for rigid links.

        Returns:
          Material creator for a rigid uniaxial material for each DOF.

        Raises:
          ValueError: If the model's dimensionality setting is
          not recognized.
        """
        # '2D Truss', '2D Frame', '3D Truss', '3D Frame'
        material_creators: dict[int, MaterialCreator]
        if self.model.dimensionality == '2D Truss':
            material_creators = {
                1: ElasticMaterialCreator(self.model, STIFF),
                2: ElasticMaterialCreator(self.model, STIFF),
            }
        elif self.model.dimensionality == '2D Frame':
            material_creators = {
                1: ElasticMaterialCreator(self.model, STIFF),
                2: ElasticMaterialCreator(self.model, STIFF),
                3: ElasticMaterialCreator(self.model, STIFF_ROT),
            }
        elif self.model.dimensionality == '3D Truss':
            material_creators = {
                1: ElasticMaterialCreator(self.model, STIFF),
                2: ElasticMaterialCreator(self.model, STIFF),
                3: ElasticMaterialCreator(self.model, STIFF),
            }
        elif self.model.dimensionality == '3D Frame':
            material_creators = {
                1: ElasticMaterialCreator(self.model, STIFF),
                2: ElasticMaterialCreator(self.model, STIFF),
                3: ElasticMaterialCreator(self.model, STIFF),
                4: ElasticMaterialCreator(self.model, STIFF_ROT),
                5: ElasticMaterialCreator(self.model, STIFF_ROT),
                6: ElasticMaterialCreator(self.model, STIFF_ROT),
            }
        else:
            msg = 'Invalid model dimensionality setting: {model.dimensionality}'
            raise ValueError(msg)
        return material_creators


@dataclass(repr=False)
class LinkCreator(BaseCreator):
    """Defines TwoNodeLink element components."""

    def add(
        self,
        tags: set[str],
        node_i: Node,
        node_j: Node,
        materials: list[UniaxialMaterial],
        directions: list[int],
        angle=0.00,
    ) -> ComponentAssembly:
        """
        Add a TwoNodeLink component.

        Returns:
          The added component.
        """
        assert node_i.uid != node_j.uid, 'Nodes need to be different.'
        component = ComponentAssembly(
            uid_generator=self.model.uid_generator,
            tags=tags,
        )
        p_i = np.array(node_i.coordinates)
        p_j = np.array(node_j.coordinates)
        axes = local_axes_from_points_and_angle(p_i, p_j, angle)  # type: ignore
        component.elements.add(
            TwoNodeLink(
                uid_generator=self.model.uid_generator,
                nodes=[node_i, node_j],
                materials=materials,
                directions=directions,
                vecyp=axes[1],
            )
        )
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        # Adding the component in the end. (It needs to have external
        # nodes before adding to the collection).
        self.model.components.add(component)
        return component


@dataclass(repr=False)
class BarCreator(BaseCreator):
    """
    Bar creator object.

    Introduces bar elements to a model.
    Bar elements are linear elements that can only carry axial load.
    """

    def add(
        self,
        tags: set[str],
        node_i: Node,
        node_j: Node,
        eo_i: numpy_array,
        eo_j: numpy_array,
        transf_type: Literal['Linear', 'Corotational'],
        area: float,
        material: UniaxialMaterial,
        outside_shape: Mesh | None = None,
        weight_per_length: float = 0.00,
    ) -> BarAssembly:
        """
        Add a bar component.

        If offsets are required, they are implemented through the
        addition of RigidLink elements.

        Returns:
          The added component.
        """
        assert node_i.uid != node_j.uid, 'Nodes need to be different.'
        # TODO(JVM): check connectivity to avoid placing another
        # element on the same pair of nodes (or warn).

        # instantiate a component assembly
        component = BarAssembly(
            uid_generator=self.model.uid_generator,
            tags=tags,
        )
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        n_i = self._prepare_connection(node_i, eo_i, component)
        n_j = self._prepare_connection(node_j, eo_j, component)

        component.elements.add(
            Bar(
                uid_generator=self.model.uid_generator,
                nodes=[n_i, n_j],
                transf_type=transf_type,
                area=area,
                material=material,
                outside_shape=outside_shape,
                weight_per_length=weight_per_length,
            )
        )

        # Adding the component in the end. (It needs to have external
        # nodes before adding to the collection).
        self.model.components.add(component)
        return component

    def _prepare_connection(
        self, node_x: Node, eo_x: numpy_array, component: BarAssembly
    ) -> Node:
        """
        Add auxiliary elements to account for offsets.

        For each end of the bar element, creates a rigid link if an
        offset exists, and returns the node to which the bar element
        should connect to. This function is called twice, once for the
        i-end and once for the j-end.  For purposes of clarity, the
        index x will be used here, meaning either i or j.

        Returns:
          A newly created node, considering the specified offset.
        """
        # if there is an offset at the x-end, create an internal node
        # and add a rigidlink element to the component assembly

        if np.linalg.norm(eo_x) < EPSILON:
            # No action needed.
            return node_x

        # '2D Truss', '2D Frame', '3D Truss', '3D Frame'
        n_x = Node(
            uid_generator=self.model.uid_generator,
            coordinates=tuple(x1 + x2 for x1, x2 in zip(node_x.coordinates, eo_x)),
        )
        component.internal_nodes.add(n_x)

        material_creators = self.define_zerolength_fixed_material_creators()
        directions, materials = ZeroLengthCreator(
            uid_generator=self.model.uid_generator,
            material_creators=material_creators,
        ).generate()

        # flip the nodes if the element is about to be defined
        # upside down
        if (
            np.allclose(
                np.array(node_x.coordinates[0:2]),
                np.array(n_x.coordinates[0:2]),
            )
            and n_x.coordinates[2] > node_x.coordinates[2]
        ):
            first_node = n_x
            second_node = node_x
        else:
            first_node = node_x
            second_node = n_x

        # Note: We don't orient the TwoNodeLink with `vecx`,
        # `vecy`, since it's simply rigid. This also avoids having
        # to separate cases for 2D/3D.
        component.elements.add(
            TwoNodeLink(
                uid_generator=self.model.uid_generator,
                nodes=[first_node, second_node],
                materials=materials,
                directions=directions,
            )
        )
        return n_x


@dataclass(repr=False)
class HingeConfig:
    """
    Configuration for `generate_hinged_component_assembly`.

    Only ZeroLength elements at the ends of the beamcolumn sequence.
    """

    zerolength_creator: ZeroLengthCreator


@dataclass(repr=False)
class HingeConfigWithBeamColumnElement(HingeConfig):
    """
    Configuration for `generate_hinged_component_assembly`.

    Two additional beamcolumn sequences, one before the first and one
    after the second ZeroLength element.
    """

    distance: float
    n_sub: int


@dataclass(repr=False)
class HingeConfigWithLink(HingeConfig):
    """
    Configuration for `generate_hinged_component_assembly`.

    Two rigid links, one before the first and one after the second
    ZeroLength element.
    """

    distance: float


@dataclass(repr=False)
class PanelZoneConfig:
    """Configuration object for panel zones."""

    doubler_plate_thickness: float
    axial_load_ratio: float
    slab_depth: float
    location: Literal['interior', 'exterior_first', 'exterior_last']
    moment_modifier: float
    consider_composite: bool


@dataclass(repr=False)
class InitialDeformationConfig:
    """
    Initial deformation configuration.

    Configuration object for initial deformations of beamcolumn
    elements defined in series.
    """

    camber_2: float
    camber_3: float
    method: Literal['sine', 'parabola']


@dataclass(repr=False)
class BeamColumnCreator(BaseCreator):
    """Introduces beamcolumn elements to a model."""

    element_type: Literal['elastic', 'disp']

    def __post_init__(self) -> None:
        """
        Code executed after initializing an object.

        Raises:
          ValueError: If an invalid element type is provided.
        """
        if self.element_type not in {'elastic', 'disp'}:
            msg = 'Invalid element type: {element_type.__name__}'
            raise ValueError(msg)

    def define_beamcolumn(
        self,
        node_i: Node,
        node_j: Node,
        offset_i: numpy_array,
        offset_j: numpy_array,
        transf_type: Literal['Linear', 'Corotational', 'PDelta'],
        section: ElasticSection | FiberSection,
        angle: float = 0.00,
        modified_stiffness_config: ModifiedStiffnessParameterConfig | None = None,
    ) -> ElasticBeamColumn | DispBeamColumn:
        """
        Define a beamcolumn element.

        Adds a beamcolumn element to the model, connecting the
        specified nodes.

        Returns:
          The added element.

        Raises:
          ValueError: If an invalid element type is provided.
        """
        p_i = np.array(node_i.coordinates) + offset_i
        p_j = np.array(node_j.coordinates) + offset_j
        axes = local_axes_from_points_and_angle(p_i, p_j, angle)  # type: ignore
        if self.element_type == 'elastic':
            assert isinstance(section, ElasticSection)
            transf = GeomTransf(
                uid_generator=self.model.uid_generator,
                transf_type=transf_type,
                offset_i=offset_i,
                offset_j=offset_j,
                x_axis=axes[0],
                y_axis=axes[1],
                z_axis=axes[2],
            )
            elm_el = ElasticBeamColumn(
                uid_generator=self.model.uid_generator,
                nodes=[node_i, node_j],
                section=section,
                geomtransf=transf,
                modified_stiffness_config=modified_stiffness_config,
            )
            res: ElasticBeamColumn | DispBeamColumn = elm_el
        elif self.element_type == 'disp':
            raise NotImplementedError
            # assert isinstance(section, FiberSection)
            # assert modified_stiffness_config is None
            # # TODO(JVM): add elastic section support
            # transf = GeomTransf(
            #     self.model.uid_generator,
            #     transf_type,
            #     offset_i,
            #     offset_j,
            #     *axes,
            # )
            # beam_integration = Lobatto(
            #     uid_generator=self.model.uid_generator,
            #     parent_section=section,
            #     n_p=2,
            # )
            # elm_disp = DispBeamColumn(
            #     uid_generator=self.model.uid_generator,
            #     parent_component=assembly,
            #     nodes=[node_i, node_j],
            #     section=section,
            #     geomtransf=transf,
            #     integration=beam_integration,
            # )
            # res = elm_disp
        else:
            msg = 'Invalid element type: {element_type.__name__}'
            raise ValueError(msg)
        return res

    def add_beamcolumn_elements_in_series(
        self,
        component: BeamColumnAssembly,
        node_i: Node,
        node_j: Node,
        eo_i: numpy_array,
        eo_j: numpy_array,
        n_sub: int,
        transf_type: Literal['Linear', 'Corotational', 'PDelta'],
        section: ElasticSection | FiberSection,
        angle: float,
        initial_deformation_config: InitialDeformationConfig | None = None,
        modified_stiffness_config: ModifiedStiffnessParameterConfig | None = None,
    ) -> None:
        """Add beamcolumn elements in series."""
        if modified_stiffness_config is not None:
            assert n_sub == 1

        num_dimensions = len(node_i.coordinates)

        if n_sub > 1:
            p_i = np.array(node_i.coordinates) + eo_i
            p_j = np.array(node_j.coordinates) + eo_j
            clear_len = np.linalg.norm(p_j - p_i)
            internal_pt_coordinates = np.linspace(
                tuple(p_i), tuple(p_j), num=n_sub + 1
            )

            if initial_deformation_config:
                t_vals = np.linspace(0.00, 1.00, num=n_sub + 1)
                if initial_deformation_config.method == 'parabola':
                    offset_vals = 4.00 * (-(t_vals**2) + t_vals)
                elif initial_deformation_config.method == 'sine':
                    offset_vals = np.sin(np.pi * t_vals)
                offset_2 = (
                    offset_vals * initial_deformation_config.camber_2 * clear_len
                )
                offset_3 = (
                    offset_vals * initial_deformation_config.camber_3 * clear_len
                )
                camber_offset: numpy_array = np.column_stack(
                    (np.zeros(n_sub + 1), offset_2, offset_3)
                )
                x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
                    p_i, p_j, angle
                )
                assert y_axis is not None
                t_glob_to_loc = transformation_matrix(x_axis, y_axis, z_axis)
                t_loc_to_glob = t_glob_to_loc.T
                camber_offset_global = (t_loc_to_glob @ camber_offset.T).T
                internal_pt_coordinates += camber_offset_global

            intnodes = []
            for i in range(1, len(internal_pt_coordinates) - 1):
                intnode = Node(
                    self.model.uid_generator,
                    (*internal_pt_coordinates[i],),
                )
                component.internal_nodes.add(intnode)
                intnodes.append(intnode)
        for i in range(n_sub):
            if i == 0:
                n_i = node_i
                o_i = eo_i
            else:
                n_i = intnodes[i - 1]
                o_i = np.zeros(num_dimensions)
            if i == n_sub - 1:
                n_j = node_j
                o_j = eo_j
            else:
                n_j = intnodes[i]
                o_j = np.zeros(num_dimensions)
            element = self.define_beamcolumn(
                node_i=n_i,
                node_j=n_j,
                offset_i=o_i,
                offset_j=o_j,
                transf_type=transf_type,
                section=section,
                angle=angle,
                modified_stiffness_config=modified_stiffness_config,
            )
            component.elements.add(element)

    def generate_plain_component_assembly(
        self,
        tags: set[str],
        node_i: Node,
        node_j: Node,
        n_sub: int,
        eo_i: numpy_array,
        eo_j: numpy_array,
        section: ElasticSection | FiberSection,
        transf_type: Literal['Linear', 'Corotational', 'PDelta'],
        angle: float = 0.00,
        initial_deformation_config: InitialDeformationConfig | None = None,
    ) -> BeamColumnAssembly:
        """
        Plain component assembly.

        Generates a plain component assembly with line elements in
        series.

        Returns:
          The created component assembly.
        """
        assert node_i.uid != node_j.uid, 'Nodes need to be different.'
        # instantiate a component assembly
        component = BeamColumnAssembly(self.model.uid_generator, tags=tags)
        # populate the component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        self.add_beamcolumn_elements_in_series(
            component=component,
            node_i=node_i,
            node_j=node_j,
            eo_i=eo_i,
            eo_j=eo_j,
            n_sub=n_sub,
            transf_type=transf_type,
            section=section,
            angle=angle,
            initial_deformation_config=initial_deformation_config,
        )

        # Adding the component in the end. (It needs to have external
        # nodes before adding to the collection).
        self.model.components.add(component)
        return component

    def generate_hinged_component_assembly(
        self,
        tags: set[str],
        node_i: Node,
        node_j: Node,
        n_sub: int,
        eo_i: numpy_array,
        eo_j: numpy_array,
        section: ElasticSection | FiberSection,
        transf_type: Literal['Linear', 'Corotational', 'PDelta'],
        angle: float,
        initial_deformation_config: InitialDeformationConfig | None = None,
        modified_stiffness_config: ModifiedStiffnessParameterConfig | None = None,
        hinge_config_i: HingeConfig
        | HingeConfigWithBeamColumnElement
        | HingeConfigWithLink
        | None = None,
        hinge_config_j: HingeConfig
        | HingeConfigWithBeamColumnElement
        | HingeConfigWithLink
        | None = None,
    ) -> BeamColumnAssembly:
        """
        Component assembly with hinges at the ends.

        Generates a component assembly that is comprised of beam-column
        elements connected in series with nonlinear springs attached
        at the ends.

        Returns:
            The defined component.
        """
        assert node_i.uid != node_j.uid, 'Nodes need to be different.'

        # Instantiate the component assembly
        component = BeamColumnAssembly(self.model.uid_generator, tags=tags)
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        p_i = np.array(node_i.coordinates) + eo_i
        p_j = np.array(node_j.coordinates) + eo_j
        x_axis, y_axis, _ = local_axes_from_points_and_angle(p_i, p_j, angle)

        def process_hinge_config(
            node: Node,
            offset: numpy_array,
            hinge_config: HingeConfig
            | HingeConfigWithBeamColumnElement
            | HingeConfigWithLink
            | None,
            *,
            is_end: bool,
        ) -> tuple[Node, numpy_array]:
            """
            Process hinge configuration for a given node.

            This method handles the hinge configuration by creating
            nodes, assigning offsets, and adding elements to the
            component assembly based on the type of hinge
            configuration provided.

            Args:
                node: The node at which the hinge is to be
                  applied.
                offset: The offset for the node's location.
                hinge_config: The hinge configuration for the
                  node. Can be one of the supported hinge types or
                  `None` if no hinge is applied.
                is_end: Whether the node is at the end of the
                  beam-column assembly.

            Returns:
                tuple
                - The connection node after processing the hinge configuration.
                - The offset to be applied to the connection node.

            Raises:
                ValueError: If an offset is specified with
                  `HingeConfig` (not possible).
                TypeError: If the provided `hinge_config` is of an unsupported type.
            """
            if hinge_config is None:
                return node, offset

            location = p_i if not is_end else p_j
            direction = x_axis if not is_end else -x_axis

            nh_out, nh_in = node, None

            if isinstance(hinge_config, HingeConfigWithBeamColumnElement):
                hinge_location = location + direction * hinge_config.distance
                nh_out = Node(
                    uid_generator=self.model.uid_generator,
                    coordinates=tuple(hinge_location),
                )
                nh_in = Node(
                    uid_generator=self.model.uid_generator,
                    coordinates=tuple(hinge_location),
                )
                component.internal_nodes.add(nh_out)
                component.internal_nodes.add(nh_in)

                self.add_beamcolumn_elements_in_series(
                    component=component,
                    node_i=node,
                    node_j=nh_out,
                    eo_i=offset,
                    eo_j=np.zeros(3),
                    n_sub=hinge_config.n_sub,
                    transf_type=transf_type,
                    section=section,
                    angle=angle,
                )
                offset = np.zeros(3)

            elif isinstance(hinge_config, HingeConfigWithLink):
                hinge_location = location + direction * hinge_config.distance
                nh_out = Node(
                    uid_generator=self.model.uid_generator,
                    coordinates=tuple(hinge_location),
                )
                nh_in = Node(
                    uid_generator=self.model.uid_generator,
                    coordinates=tuple(hinge_location),
                )
                component.internal_nodes.add(nh_out)
                component.internal_nodes.add(nh_in)

                material_creators = self.define_zerolength_fixed_material_creators()
                directions, materials = ZeroLengthCreator(
                    uid_generator=self.model.uid_generator,
                    material_creators=material_creators,
                ).generate()

                component.elements.add(
                    TwoNodeLink(
                        uid_generator=self.model.uid_generator,
                        nodes=[node, nh_out],
                        materials=materials,
                        directions=directions,
                    )
                )
                offset = np.zeros(3)

            elif isinstance(hinge_config, HingeConfig):
                if np.linalg.norm(offset) > EPSILON:
                    msg = "Can't have offset with `HingeConfig`. Use `HingeConfigWithElement`."
                    raise ValueError(msg)
                hinge_location = location
                nh_out = node
                nh_in = Node(
                    uid_generator=self.model.uid_generator,
                    coordinates=tuple(hinge_location),
                )
                component.internal_nodes.add(nh_in)
            else:
                msg = f'Invalid hinge_config type: {type(hinge_config)}'
                raise TypeError(msg)

            component.elements.add(
                hinge_config.zerolength_creator.define_element(
                    node_i=nh_out,
                    node_j=nh_in,
                    x_axis=direction,
                    y_axis=y_axis,
                )
            )

            if nh_in is not None:
                nh_return = nh_in
            else:
                nh_return = nh_out

            return nh_return, offset

        # Process hinge configurations for node_i and node_j
        conn_node_i, conn_eo_i = process_hinge_config(
            node_i, eo_i, hinge_config_i, is_end=False
        )
        conn_node_j, conn_eo_j = process_hinge_config(
            node_j, eo_j, hinge_config_j, is_end=True
        )

        # Add the final beam-column elements
        self.add_beamcolumn_elements_in_series(
            component=component,
            node_i=conn_node_i,
            node_j=conn_node_j,
            eo_i=conn_eo_i,
            eo_j=conn_eo_j,
            n_sub=n_sub,
            transf_type=transf_type,
            section=section,
            angle=angle,
            initial_deformation_config=initial_deformation_config,
            modified_stiffness_config=modified_stiffness_config,
        )

        # Adding the component in the end. (It needs to have external
        # nodes before adding to the collection).
        self.model.components.add(component)
        return component



"""Determine the latest version of the osmg package on PyPI."""

import argparse

import requests


def get_latest_version(package_name: str) -> str:
    """
    Determine the latest version of the osmg package on PyPI.

    Returns:
      The latest version.
    """
    url = f'https://pypi.org/pypi/{package_name}/json'

    response = requests.get(url, timeout=10)
    package_info = response.json()
    return str(package_info['info']['version'])


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--package_name', default='osmg', help='Name of the package.'
    )
    args = parser.parse_args()

    package = args.package_name
    version = get_latest_version(package)
    print(version)  # noqa: T201



"""Collections are designated containers of objects of a particular type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar, Literal

import numpy as np

from osmg.analysis.common import UDL
from osmg.core import common
from osmg.core.uid_object import UIDObject
from osmg.geometry.transformations import (
    transformation_matrix,
    transformation_matrix_2d,
)
from osmg.model_objects.element import Bar, BeamColumnElement, Element
from osmg.model_objects.node import Node

TV = TypeVar('TV')

if TYPE_CHECKING:
    from osmg.core.common import numpy_array
    from osmg.model_objects import element
    from osmg.model_objects.section import Section


@dataclass(repr=False)
class Collection(dict[int, TV]):
    """Collection of objects."""

    def add(self, obj: TV) -> None:
        """
        Add an object to the collection.

        The object needs to have a unique id attribute, `uid`.

        Arguments:
          obj: Object to be added.

        Raises:
          ValueError: If the object does not have a uid attribute.
          ValueError: If the object already exists.
        """
        assert hasattr(obj, 'uid')
        if obj.uid in self:
            msg = f'uid {obj.uid} already exists'
            raise ValueError(msg)
        self[obj.uid] = obj

    def retrieve_by_attr(self, attr: str, val: object) -> TV:
        """
        Retrieve object by attribute value.

        Retrieve an object from the collection based on an attribute
        value. If more than one instances satisfy the criterion, the
        function returns the first occurrence.

        Arguments:
          attr: The name of the attribute.
          val: The value that the attribute should have.

        Returns:
          The retrieved object.

        Raises:
          ValueError: If the object is not found.
        """
        res = None
        for thing in self.values():
            found = False
            if hasattr(thing, attr):
                other_val = getattr(thing, attr)
                if other_val == val:
                    res = thing
                    found = True
                    break
            if found:
                break
        if res is None:
            msg = (
                f'Item having the value "{val}" '
                f'in the attribute "{attr}" not found in collection.'
            )
            raise ValueError(msg)
        return res

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'Collection Object\n'
        res += f'ID: {id(self)}\n'
        res += f'Number of objects: {len(self)}\n'
        return res


@dataclass(repr=False)
class NamedCollection(Collection[TV], Generic[TV]):
    """
    A collection that allows assigning and retrieving objects by name.

    Attributes:
    ----------
    named_contents : dict[str, TV]
        A mapping of names to objects for easy retrieval.
    """

    named_contents: dict[str, TV] = field(default_factory=dict)

    def add(self, obj: TV, name: str | None = None) -> None:
        """
        Add a named object to the collection.

        The Node needs to have a unique id attribute, `uid`.
        Parameters
        ----------
        obj : TV
            Object to be added.
        name : str
            The name to assign to the object.

        """
        super().add(obj)
        if name:
            self.named_contents[name] = obj

    def retrieve_by_name(self, name: str) -> TV:
        """
        Retrieve an object by its name.

        Params:
          name: The name of the object to retrieve.

        Returns:
            The object associated with the given name.

        Raises:
          KeyError: If no object is found with the given name.
        """
        if name not in self.named_contents:
            msg = f"No object found with the name '{name}'."
            raise KeyError(msg)
        return self.named_contents[name]

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
        -------
        str
            The string representation of the object.
        """
        res = ''
        res += 'NamedCollection Object\n'
        res += f'ID: {id(self)}\n'
        res += f'Number of objects: {len(self)}\n'
        res += f'Named objects: {list(self.named_contents.keys())}\n'
        return res


@dataclass(repr=False)
class ElementCollection(Collection[Element]):
    """Collection of elements."""

    def add(self, obj: Element) -> None:
        """
        Add an element to the collection.

        Arguments:
          obj: Object to be added.
        """
        super().add(obj)


@dataclass(repr=False)
class NodeCollection(NamedCollection[Node]):
    """
    A collection specifically for managing named Nodes.

    Methods:
    -------
    search_by_coordinates:
        Search for a node at the specified coordinates.
    """

    def search_by_coordinates(
        self, coordinates: tuple[float, ...], radius: float = common.EPSILON
    ) -> Node | None:
        """
        Obtain the node that occupies a given point if it exists.

        Params:
          coordinates: The coordinates to search at.

        Returns:
          Node: The node at the given coordinates, or None if no node is found.
        """
        candidate_pt: numpy_array = np.array(coordinates)
        for other_node in self.values():
            other_pt: numpy_array = np.array(other_node.coordinates)
            if np.linalg.norm(candidate_pt - other_pt) < radius:
                return other_node
        return None

    def search_by_coordinates_or_raise(
        self, coordinates: tuple[float, ...], radius: float = common.EPSILON
    ) -> Node:
        """
        Obtain the node that occupies a given point if it exists.

        Raise an error if it does not exist.

        Params:
          coordinates: The coordinates to search at.

        Returns:
          The node at the given coordinates, or None if no node is found.

        Raises:
          ValueError: If a node is not found.
        """
        node = self.search_by_coordinates(coordinates=coordinates, radius=radius)
        if node is None:
            msg = f'No node found at given coordinates: {coordinates}.'
            raise ValueError(msg)
        return node


@dataclass
class ComponentAssembly(UIDObject):
    """
    Component assembly object.

    A component assembly represents some part of a structure and holds
    various lower-level elements such as nodes and beam-column
    elements.

    Attributes:
        uid: Unique identifier of the component assembly.
        tags: List of tags, used to filter component assemblies if
            needed.
        external_nodes: The external nodes to which the component
            assembly is connected.
        internal_nodes: Internal nodes that are required for the
            connectivity of the elements within the component
            assembly. These nodes only exist as part of the component
            assembly.
        elements: Collection containing the elements that are part of
            the component assembly.
    """

    tags: set[str]
    external_nodes: NodeCollection = field(default_factory=NodeCollection)
    internal_nodes: NodeCollection = field(default_factory=NodeCollection)
    elements: ElementCollection = field(default_factory=ElementCollection)

    def __hash__(self) -> int:
        """Return the hash of the object based on its UID."""
        return hash(self.uid)

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'Component assembly object\n'
        res += f'uid: {self.uid}\n'
        res += 'External Nodes\n'
        for node in self.external_nodes.values():
            res += f'  {node.uid}, {node.coordinates}'
        return res

    def element_connectivity(
        self,
    ) -> dict[tuple[int, ...], element.Element]:
        """
        Element connectivity.

        Returns:
          The connectivity of all elements. Elements are connected to
          external nodes. Each component assembly can be represented
          by a tuple of node uids of its connected nodes in ascending
          order. This method returns a dictionary having these tuples
          as keys, and the associated components as values.
        """
        res = {}
        elms = self.elements.values()
        for elm in elms:
            uids = [x.uid for x in elm.nodes]
            uids.sort()
            uids_tuple = (*uids,)
            res[uids_tuple] = elm
        return res


@dataclass
class BarAssembly(ComponentAssembly):
    """Component assembly for bar elements."""


    def __hash__(self) -> int:
        """Return the hash of the object based on its UID."""
        return hash(self.uid)

    def __eq__(self, other: object) -> bool:
        """
        Check equality based on the UID.

        Returns:
          True if it is equal, False otherwise.
        """
        if not isinstance(other, UIDObject):
            return False
        return self.uid == other.uid


@dataclass
class BeamColumnAssembly(ComponentAssembly):
    """
    Beamcolumn assembly object.

    A beamcolumn assembly is a collection meant to represent beams and
    columns. It is modeled with beamcolumn elements connected in
    series, with optional zerolength elements at certain
    locations. All beamcolumn elements in the collection need to be
    colinear, since parts of osmg rely on this assumption.
    """

    def clear_length(self) -> float:
        """
        Get the clear length of the component assembly.

        Returns:
          The clear length.
        """
        beamcolumn_elements = [
            element
            for element in self.elements.values()
            if isinstance(element, BeamColumnElement)
        ]
        first_element = beamcolumn_elements[0]
        last_element = beamcolumn_elements[-1]

        p_i = (
            np.array(first_element.nodes[0].coordinates)
            + first_element.geomtransf.offset_i
        )
        p_j = (
            np.array(last_element.nodes[1].coordinates)
            + last_element.geomtransf.offset_j
        )
        return float(np.linalg.norm(p_i - p_j))

    def calculate_element_udl(self, udl: UDL) -> dict[int, UDL]:
        """
        Distribute the given UDL to the beamcolumn elements.

        Given a UDL, determine the elements meant to take the load and
        return what is necessary to define the load in OpenSees: The
        tags of the elements and the load expressed in the local
        coordinate system of each element.

        Returns:
          Dictionary mapping the UID of the beamcolumn elements with
          UDL values expressed in the members' local coordinate
          system.
        """
        local_udls: dict[int, UDL] = {}
        for beamcolumn_element in self.elements.values():
            if not isinstance(beamcolumn_element, BeamColumnElement):
                continue
            if beamcolumn_element.geomtransf.y_axis is None:
                transformation_mat = transformation_matrix_2d(
                    beamcolumn_element.geomtransf.x_axis,
                    beamcolumn_element.geomtransf.z_axis,
                )
            else:
                transformation_mat = transformation_matrix(
                    beamcolumn_element.geomtransf.x_axis,
                    beamcolumn_element.geomtransf.y_axis,
                    beamcolumn_element.geomtransf.z_axis,
                )
            udl_local = UDL(transformation_mat @ np.array(udl))
            local_udls[beamcolumn_element.uid] = udl_local
        return local_udls

    def get_section(self) -> Section:
        """
        Retrieve the section used by beamcolumn elements.

        Raises:
          ValueError: If not all beamcolumn elements share the same
            section.
          ValueError: If there are no beamcolumn elements in the
            component.

        Returns:
          The common section.
        """
        sections = [
            element.section
            for element in self.elements.values()
            if isinstance(element, BeamColumnElement)
        ]
        if not sections:
            msg = 'No elements.'
            raise ValueError(msg)
        if len({section.name for section in sections}) > 1:
            msg = 'Multiple sections found.'
            raise ValueError(msg)
        return sections[0]

    def update_section(self, section: Section) -> None:
        """Update the section of all internal elements."""
        for element in self.elements.values():
            if isinstance(element, BeamColumnElement):
                element.section = section

    def __hash__(self) -> int:
        """Return the hash of the object based on its UID."""
        return hash(self.uid)

    def __eq__(self, other: object) -> bool:
        """
        Check equality based on the UID.

        Returns:
          True if it is equal, False otherwise.
        """
        if not isinstance(other, UIDObject):
            return False
        return self.uid == other.uid


@dataclass(repr=False)
class ComponentAssemblyCollection(Collection[ComponentAssembly]):
    """
    Collection of component assemblies with connectivity tracking.

    Attributes:
    ----------
    connectivity_map : dict[tuple[int, ...], int]
        A dictionary where the keys are sorted tuples of external node
        uids, and the values are the uids of the corresponding
        component assemblies.
    """

    connectivity_map: dict[tuple[int, ...], int] = field(default_factory=dict)

    def add(self, obj: ComponentAssembly) -> None:
        """
        Add a component to the collection and update connectivity information.

        Arguments:
          obj: ComponentAssembly to be added.

        Raises:
          ValueError: If the object is missing required attributes or
                      if connectivity information conflicts.
        """
        super().add(obj)

        # Ensure external nodes have unique IDs
        if not obj.external_nodes:
            msg = (
                'ComponentAssembly must have external nodes to update connectivity.'
            )
            raise ValueError(msg)

        # Extract external node uids
        external_uids = [node.uid for node in obj.external_nodes.values()]
        if not external_uids:
            msg = 'External nodes must have valid unique IDs.'
            raise ValueError(msg)

        # Create a sorted tuple of external node uids
        sorted_uids = tuple(sorted(external_uids))

        # Update connectivity map
        if sorted_uids in self.connectivity_map:
            msg = (
                f'Connectivity information for nodes {sorted_uids} '
                f'already exists for ComponentAssembly {self.connectivity_map[sorted_uids]}.'
            )
            raise ValueError(msg)

        self.connectivity_map[sorted_uids] = obj.uid

    def search_by_nodes(self, nodes: list[Node]) -> ComponentAssembly | None:
        """
        Search and return a component assembly connected to the given nodes.

        Arguments:
          nodes: List of Node objects to check for connectivity.

        Returns:
          The connected ComponentAssembly if found, else None.
        """
        # Extract sorted uids from the given nodes
        node_uids = tuple(sorted(node.uid for node in nodes))

        # Check if the connectivity map contains the sorted uids
        if node_uids in self.connectivity_map:
            assembly_uid = self.connectivity_map[node_uids]
            return self[assembly_uid]

        return None

    def search_by_nodes_or_raise(self, nodes: list[Node]) -> ComponentAssembly:
        """
        Search and return a component assembly connected to the given nodes.

        Raise an error if not found.

        Arguments:
          nodes: List of Node objects to check for connectivity.

        Returns:
          The connected ComponentAssembly.

        Raises:
          ValueError: If not found
        """
        component = self.search_by_nodes(nodes)
        if component is None:
            msg = 'Component not found.'
            raise ValueError(msg)
        return component

    def search_by_axis_aligned_bounding_box(
        self,
        bounding_box: tuple[tuple[float, float], ...],
        *,
        offset: float = 1e-5,
        exclusive: bool = True,
    ) -> list[ComponentAssembly]:
        """
        Search by axis-aligned bounding box.

        Search and return all component assemblies connected to nodes
        within the specified bounding box.

        Arguments:
          bounding_box: A tuple defining the bounding box in the form:
                        ((xmin, xmax), (ymin, ymax), optionally (zmin,
                        zmax)).
          offset: Expand the boudning box to combat numerical
                  precision issues.
          exclusive: Whether to return only components connecting
                     exclusively to the selected nodes (default:
                     True).

        Returns:
          A list of ComponentAssemblies connected to the selected
          nodes.
        """
        selected_uids = self._filter_nodes_by_bounding_box(bounding_box, offset)
        return self._find_assemblies_by_uids(selected_uids, exclusive=exclusive)

    def search_by_axis_projection(
        self,
        point: tuple[float, float, float],
        axes: Literal['xy', 'xz', 'yz'],
        *,
        offset: float = 1e-5,
        exclusive: bool = True,
    ) -> list[ComponentAssembly]:
        """
        Search by infinite axis-aligned bounding box.

        Search and return all component assemblies connected to nodes
        within an infinite bounding box projected in the specified
        axes.

        Arguments:
          point: A tuple defining the coordinates of the point.
          axes: Axes of the bounding box projection ('xy', 'xz', or
               'yz').
          offset: Small offset to add to the bounding box (default:
                  1e-5).
          exclusive: Whether to return only components connecting
                     exclusively to the selected nodes (default:
                     True).

        Returns:
          A list of ComponentAssemblies connected to the selected
          nodes.

        Raises:
          ValueError: If the axes argument is invalid.
        """
        axes_indices = {'xy': (0, 1), 'xz': (0, 2), 'yz': (1, 2)}

        if axes not in axes_indices:
            msg = 'Invalid axes value. Must be one of `xy`, `xz`, or `yz`.'
            raise ValueError(msg)

        selected_indices = axes_indices[axes]
        bounding_box = [
            (-np.inf, np.inf)
            if i not in selected_indices
            else (point[i] - offset, point[i] + offset)
            for i in range(len(point))
        ]

        selected_uids = self._filter_nodes_by_bounding_box(
            tuple(bounding_box), offset
        )
        return self._find_assemblies_by_uids(selected_uids, exclusive=exclusive)

    def _filter_nodes_by_bounding_box(
        self, bounding_box: tuple[tuple[float, float], ...], offset: float
    ) -> set[int]:
        """
        Filter nodes by a bounding box.

        Arguments:
          bounding_box: A tuple defining the bounding box.
          offset: Expand the boudning box to combat numerical
                  precision issues.

        Returns:
          A set of UIDs for nodes within the bounding box.

        Raises:
          ValueError: If the dimensionality of the bounding box is
          invalid.
        """
        all_nodes = self._get_all_nodes()
        if not all_nodes:
            return set()
        some_node = all_nodes[0]
        ndf = len(some_node.coordinates)

        if len(bounding_box) != ndf:
            msg = (
                f'Invalid bounding box dimensionality '
                f'({len(bounding_box)}), should be {ndf}.'
            )
            raise ValueError(msg)

        return {
            node.uid
            for node in all_nodes
            if all(
                bounds[0] - offset <= node.coordinates[i] <= bounds[1] + offset
                for i, bounds in enumerate(bounding_box)
            )
        }

    def _find_assemblies_by_uids(
        self,
        selected_uids: set[int],
        *,
        exclusive: bool,
    ) -> list[ComponentAssembly]:
        """
        Find component assemblies by UIDs.

        Arguments:
          selected_uids: A set of node UIDs to match.
          exclusive: Whether to return only components connecting
                     exclusively to the selected nodes.

        Returns:
          A list of ComponentAssemblies connected to the selected
          nodes.
        """
        assemblies = []
        for uids, assembly_uid in self.connectivity_map.items():
            uids_set = set(uids)
            if uids_set.issubset(selected_uids) and exclusive:
                # Include only if the assembly connects exclusively to
                # these nodes
                assemblies.append(self[assembly_uid])
            if uids_set.intersection(selected_uids) and not exclusive:
                # Include if the assembly connects to these nodes (and
                # possibly others)
                assemblies.append(self[assembly_uid])
        return assemblies

    def _get_all_nodes(self) -> list[Node]:
        """
        Retrieve all nodes in the collection.

        Returns:
          A list of all Node objects in the collection.
        """
        all_nodes = []
        for assembly in self.values():
            all_nodes.extend(assembly.external_nodes.values())
        return all_nodes

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'ComponentAssemblyCollection Object\n'
        res += f'ID: {id(self)}\n'
        res += f'Number of assemblies: {len(self)}\n'
        res += f'Connectivity map: {self.connectivity_map}\n'
        return res



"""objects that create sections."""

from __future__ import annotations

import importlib.resources
import json
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel

from osmg.geometry.mesh_shapes import w_mesh
from osmg.model_objects.section import ElasticSection

if TYPE_CHECKING:
    from osmg.creators.uid import UIDGenerator


# Base class for sections
class BaseSectionData(BaseModel):
    """Base model for section data."""

    Type: str
    A: float
    Ix: float
    Iy: float
    J: float
    W: float


# W Section model
class WSectionData(BaseSectionData):
    """Model for W sections."""

    bf: float
    d: float
    tw: float
    tf: float
    Zx: float
    Zy: float
    rx: float
    ry: float


# HSS Section model
class HSSSectionData(BaseSectionData):
    """Model for HSS sections."""

    t: float
    D: float
    B: float


# Union for supported section types
SectionData = WSectionData | HSSSectionData


class SectionDatabase(BaseModel):
    """Model for the entire section database."""

    sections: dict[str, SectionData]


class AISC_Database_Section_Creator:
    """
    Create frame member sections from a predefined database.

    Force units are `lb` and length units are `in`.
    """

    def __init__(
        self, uid_generator: UIDGenerator, database_path: str | None = None
    ) -> None:
        """Instantiate object."""
        self.uid_generator = uid_generator
        self.database_path = database_path
        self._load_database()

    def _load_database(self) -> None:
        """Load and validate the JSON database."""
        if self.database_path is None:
            with importlib.resources.open_text('osmg.data', 'sections.json') as f:
                raw_data = f.read()
        else:
            with Path(self.database_path).open('r', encoding='utf-8') as f:
                raw_data = f.read()

        json_data = json.loads(raw_data)
        valid_sections: dict[str, SectionData] = {}
        unsupported_types = []
        for name, section in json_data.items():
            section_type = section['Type']
            if section_type == 'W':
                valid_sections[name] = WSectionData(**section)
            # elif section_type == 'HSS':
            #     valid_sections[name] = HSSSectionData(**section)
            else:
                unsupported_types.append(section_type)
        if unsupported_types:
            print(f'Skipping unsupported section types: {set(unsupported_types)}')  # noqa: T201

        self.section_database = SectionDatabase(sections=valid_sections)

    def get_available(self) -> list[str]:
        """
        Determine which sections are available in the database.

        Returns:
          List of available sections.
        """
        return list(self.section_database.sections.keys())

    def load_elastic_section(
        self, section_label: str, e_modulus: float, g_modulus: float
    ) -> ElasticSection:
        """
        Load a section from the database.

        Args:
            section_label: Label of the section to load.
            e_modulus: Elastic modulus.
            g_modulus: Shear modulus.

        Returns:
            ElasticSection: The section.

        Raises:
            ValueError: If the provided `section_label` is unavailable.
            NotImplementedError: If the section type is unsupported.
        """
        section_data = self.section_database.sections.get(section_label)

        if section_data is None:
            msg = f'The provided `section_label` is unavailable: {section_label}.'
            raise ValueError(msg)

        if isinstance(section_data, WSectionData):
            # Geometry calculations for W sections
            outside_shape = w_mesh(
                section_data.bf,
                section_data.d,
                section_data.tw,
                section_data.tf,
                section_data.A,
            )
            bbox = outside_shape.bounding_box()
            z_min, y_min, z_max, y_max = bbox.flatten()
            snap_points = {
                'centroid': np.array([0.0, 0.0]),
                'top_center': np.array([0.0, -y_max]),
                'top_left': np.array([-z_min, -y_max]),
                'top_right': np.array([-z_max, -y_max]),
                'center_left': np.array([-z_min, 0.0]),
                'center_right': np.array([-z_max, 0.0]),
                'bottom_center': np.array([0.0, -y_min]),
                'bottom_left': np.array([-z_min, -y_min]),
                'bottom_right': np.array([-z_max, -y_min]),
            }

            return ElasticSection(
                self.uid_generator,
                section_label,
                e_modulus,
                section_data.A,
                section_data.Iy,
                section_data.Ix,
                g_modulus,
                section_data.J,
                section_data.W / 12.00,  # lb/in
                outside_shape,
                snap_points,
                properties=section_data,
            )
        msg = f'Section type is unsupported: {section_data.Type}'
        raise NotImplementedError(msg)



"""Unit tests for coordinate transformation operations."""

import numpy as np
import pytest

from osmg.geometry.transformations import (
    local_axes_from_points_and_angle,
    offset_transformation_3d,
    rotation_matrix_2d,
    rotation_matrix_3d,
    transformation_matrix,
)


class TestRotationMatrix2D:
    """Tests for the `rotation_matrix_2d` function."""

    def test_rotation_matrix_2d(self) -> None:
        """Test the 2D rotation matrix."""
        result = rotation_matrix_2d(np.pi / 2)
        expected = np.array([[0.0, -1.0], [1.0, 0.0]])
        assert np.allclose(result, expected)

    def test_rotation_matrix_2d_invalid_input(self) -> None:
        """Test that a TypeError is raised for invalid input."""
        with pytest.raises(TypeError, match='ang parameter should be a float.'):
            rotation_matrix_2d('90')  # type: ignore  # Invalid input type


class TestRotationMatrix3D:
    """Tests for the `rotation_matrix_3d` function."""

    def test_rotation_matrix_3d(self) -> None:
        """Test the 3D rotation matrix."""
        axis = np.array([1, 0, 0])
        theta = np.pi / 2
        result = rotation_matrix_3d(axis, theta)
        expected = np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, 0.0, -1.0],
                [0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(result, expected)

    def test_rotation_matrix_3d_invalid_axis(self) -> None:
        """Test that the function handles invalid axis input."""
        axis = np.array([0, 0])  # Invalid axis (not 3D)
        theta = np.pi / 2
        with pytest.raises(ValueError, match='not enough values to unpack'):
            rotation_matrix_3d(axis, theta)


class TestTransformationMatrix:
    """Tests for the `transformation_matrix` function."""

    def test_transformation_matrix_identity(self) -> None:
        """Test the transformation matrix for identity axes."""
        vec_x = np.array([1.0, 0.0, 0.0])
        vec_y = np.array([0.0, 1.0, 0.0])
        vec_z = np.array([0.0, 0.0, 1.0])
        result = transformation_matrix(vec_x, vec_y, vec_z)
        expected = np.eye(3)
        assert np.allclose(result, expected)

    def test_transformation_matrix_permuted_axes(self) -> None:
        """Test the transformation matrix with permuted axes."""
        vec_x = np.array([1.0, 0.0, 0.0])
        vec_y = np.array([0.0, 0.0, 1.0])
        vec_z = np.array([0.0, 1.0, 0.0])
        result = transformation_matrix(vec_x, vec_y, vec_z)
        expected = np.array([[1.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0]])
        assert np.allclose(result, expected)


class TestLocalAxesFromPointsAndAngle:
    """Tests for the `local_axes_from_points_and_angle` function."""

    def test_local_axes_horizontal(self) -> None:
        """Test local axes for a horizontal element."""
        point_i = np.array([0.0, 0.0, 0.0])
        point_j = np.array([1.0, 0.0, 0.0])
        ang = 0.0
        result = local_axes_from_points_and_angle(point_i, point_j, ang)
        expected_x = np.array([1.0, 0.0, 0.0])
        expected_y = np.array([0.0, 0.0, 1.0])
        expected_z = np.array([0.0, -1.0, 0.0])
        assert np.allclose(result[0], expected_x)  # type: ignore
        assert np.allclose(result[1], expected_y)  # type: ignore
        assert np.allclose(result[2], expected_z)  # type: ignore

    def test_local_axes_vertical(self) -> None:
        """Test local axes for a vertical element."""
        point_i = np.array([0.0, 0.0, 0.0])
        point_j = np.array([0.0, 0.0, 1.0])
        ang = np.pi / 4
        result = local_axes_from_points_and_angle(point_j, point_i, ang)
        expected_x = np.array([0.0, 0.0, -1.0])
        expected_y = np.array([-np.sqrt(2) / 2, np.sqrt(2) / 2, 0.0])
        expected_z = np.array([np.sqrt(2) / 2, np.sqrt(2) / 2, 0.0])
        assert np.allclose(result[0], expected_x)  # type: ignore
        assert np.allclose(result[1], expected_y)  # type: ignore
        assert np.allclose(result[2], expected_z)  # type: ignore

    def test_local_axes_vertical_upside_down(self) -> None:
        """Test that a ValueError is raised for an upside-down vertical element."""
        point_i = np.array([0.0, 0.0, 0.0])
        point_j = np.array([0.0, 0.0, 1.0])
        ang = 0.0
        with pytest.raises(ValueError, match='Vertical element defined upside down'):
            local_axes_from_points_and_angle(point_i, point_j, ang)


class TestOffsetTransformation:
    """Tests for the `offset_transformation` function."""

    def test_offset_transformation(self) -> None:
        """Test the offset transformation calculation."""
        offset = np.array([1.0, 0.0, 0.0])
        u_vec = np.array([0.01, -0.02, 0.005])
        r_vec = np.array([0.0002, -0.0003, 0.0001])
        result = offset_transformation_3d(offset, u_vec, r_vec)
        expected = np.array([0.01, -0.0199, 0.0053])
        assert np.allclose(result, expected)  # type: ignore



"""Unit tests for load case and related functions."""

import numpy as np
import pandas as pd
import pytest

from osmg.analysis.load_case import (
    combine,
    combine_single,
    ensure_minmax_level_exists_or_add,
)


class TestAddMinMaxLevel:
    """Tests the `ensure_minmax_level_exists_or_add` function."""

    def test_ensure_minmax_level_exists_or_add_single_level(self) -> None:
        """Test adding 'min/max' level to a single-level MultiIndex DataFrame."""
        test_df = pd.DataFrame(
            [[1, 2], [3, 4]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )
        result = ensure_minmax_level_exists_or_add(test_df)

        expected_columns = pd.MultiIndex.from_tuples(
            [
                ('a', 'x', 'max'),
                ('a', 'x', 'min'),
                ('a', 'y', 'max'),
                ('a', 'y', 'min'),
            ],
            names=['level_1', 'level_2', 'min/max'],
        )
        assert result.columns.equals(expected_columns)
        assert np.allclose(result.iloc[:, :2].to_numpy(), test_df.to_numpy())

    def test_ensure_minmax_level_exists_or_add_existing_level(self) -> None:
        """Does not alter a DataFrame with an existing 'min/max' level."""
        has_minmax = pd.DataFrame(
            [[1, 2], [3, 4]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x', 'max'), ('a', 'y', 'min')],
                names=['level_1', 'level_2', 'min/max'],
            ),
        )
        result = ensure_minmax_level_exists_or_add(has_minmax)
        pd.testing.assert_frame_equal(result, has_minmax)


class TestCombineSingle:
    """Tests the `combine_single` function."""

    def setup_method(self) -> None:
        """Set up common test data."""
        self.df1 = pd.DataFrame(
            [[1, 2], [3, 4]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )
        self.df2 = pd.DataFrame(
            [[5, 6], [7, 8]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )

    def test_combine_add(self) -> None:
        """Test the 'add' action in combine."""
        result = combine_single(self.df1, self.df2, 'add')
        expected = pd.DataFrame(
            [[6, 8], [10, 12]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')],
                names=['level_1', 'level_2'],
            ),
        )
        pd.testing.assert_frame_equal(result, expected)

    def test_combine_envelope(self) -> None:
        """Test the 'envelope' action in combine."""
        result = combine_single(self.df1, self.df2, 'envelope')
        expected = pd.DataFrame(
            [[5, 5, 2, 2], [7, 7, 4, 4]],
            columns=pd.MultiIndex.from_tuples(
                [
                    ('a', 'x', 'max'),
                    ('a', 'x', 'min'),
                    ('a', 'y', 'max'),
                    ('a', 'y', 'min'),
                ],
                names=['level_1', 'level_2', 'min/max'],
            ),
        )
        pd.testing.assert_frame_equal(result, expected)

    def test_combine_invalid_action(self) -> None:
        """An invalid action raises a ValueError."""
        with pytest.raises(
            ValueError, match='Action must be one of `add` or `envelope`.'
        ):
            combine_single(self.df1, self.df2, 'invalid')  # type: ignore

    def test_combine_different_columns(self) -> None:
        """Combining DataFrames with different columns raises an error."""
        df3 = pd.DataFrame(
            [[9], [10]],  # Adjust to match the number of columns
            columns=pd.MultiIndex.from_tuples(
                [('b', 'z')], names=['level_1', 'level_2']
            ),
        )
        with pytest.raises(
            ValueError, match='Cannot align DataFrames with different columns'
        ):
            combine_single(self.df1, df3, 'add')


class TestCombine:
    """Tests the `combine` function."""

    def setup_method(self) -> None:
        """Set up common test data."""
        self.df1 = pd.DataFrame(
            [[1, 2], [3, 4]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )
        self.df2 = pd.DataFrame(
            [[5, 6], [7, 8]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )
        self.df3 = pd.DataFrame(
            [[9, 10], [11, 12]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')], names=['level_1', 'level_2']
            ),
        )

    def test_combine_add(self) -> None:
        """Test combining  DataFrames with the 'add' action."""
        result = combine([self.df1, self.df2, self.df3], 'add')
        expected = pd.DataFrame(
            [[15, 18], [21, 24]],
            columns=pd.MultiIndex.from_tuples(
                [('a', 'x'), ('a', 'y')],
                names=['level_1', 'level_2'],
            ),
        )
        pd.testing.assert_frame_equal(result, expected)

    def test_combine_envelope(self) -> None:
        """Test combining  DataFrames with the 'envelope' action."""
        result = combine([self.df1, self.df2, self.df3], 'envelope')
        expected = pd.DataFrame(
            [[9, 9, 5, 2], [11, 11, 7, 4]],
            columns=pd.MultiIndex.from_tuples(
                [
                    ('a', 'x', 'max'),
                    ('a', 'x', 'min'),
                    ('a', 'y', 'max'),
                    ('a', 'y', 'min'),
                ],
                names=['level_1', 'level_2', 'min/max'],
            ),
        )
        pd.testing.assert_frame_equal(result, expected)

    def test_combine_insufficient_dataframes(self) -> None:
        """Test that combining fewer than two DataFrames raises a ValueError."""
        with pytest.raises(
            ValueError, match='At least two DataFrames are required to combine.'
        ):
            combine([self.df1], 'add')

    def test_combine_different_columns(self) -> None:
        """Test combining  DataFrames with mismatched columns raises an error."""
        df4 = pd.DataFrame(
            [[1], [2]],
            columns=pd.MultiIndex.from_tuples(
                [('b', 'z')], names=['level_1', 'level_2']
            ),
        )
        with pytest.raises(
            ValueError, match='Cannot align DataFrames with different columns'
        ):
            combine([self.df1, self.df2, df4], 'add')

    def test_combine_invalid_action(self) -> None:
        """Test that an invalid action raises a ValueError."""
        with pytest.raises(
            ValueError, match='Action must be one of `add` or `envelope`.'
        ):
            combine([self.df1, self.df2], 'invalid')  # type: ignore



"""Common definitions."""

from __future__ import annotations

import re
from pprint import pprint
from typing import Hashable, OrderedDict, TypeVar

import numpy as np
import numpy.typing as npt

# very big, very small numbers used for
# comparing floats and hashing
EPSILON = 1.00e-6
ALPHA = 1.00e8

# gravitational acceleration
G_CONST_IMPERIAL = 386.22  # in/s**2
G_CONST_SI = 9.81  # m/s**2

# quantities to use for extreme stiffnesses
STIFF_ROT = 1.0e15
STIFF = 1.0e10
TINY = 1.0e-12

NDM: dict[str, int] = {
    '1D1DOF': 1,
    '2D Truss': 2,
    '2D Frame': 2,
    '3D Truss': 3,
    '3D Frame': 3,
}
NDF: dict[str, int] = {
    '1D1DOF': 1,
    '2D Truss': 2,
    '2D Frame': 3,
    '3D Truss': 3,
    '3D Frame': 6,
}

TWO_DIMENSIONAL = 2
THREE_DIMENSIONAL = 3

numpy_array = npt.NDArray[np.float64]


def methods(obj: object) -> list[str]:
    """
    Get the methods of an object.

    Returns:
      The names of all methods of an object, excluding the dunder
      methods.

    Example:
        >>> class TestClass:
        ...     def method_1(self):
        ...         pass
        ...
        ...     def method_2(self):
        ...         pass
        ...
        >>> obj = TestClass()
        >>> methods(obj)
        ['method_1', 'method_2']
    """
    object_methods = [
        method_name
        for method_name in dir(obj)
        if callable(getattr(obj, method_name))
    ]
    pattern = r'__.*__'
    return [s for s in object_methods if not re.match(pattern, s)]


def print_methods(obj: object) -> None:
    """Print the methods of an object."""
    object_methods = methods(obj)
    pprint(object_methods)  # noqa: T203


def print_dir(obj: object) -> None:
    """Print the entire output of `dir()` of an object."""
    pprint(dir(obj))  # noqa: T203


K = TypeVar('K', bound=Hashable)  # Represents the key type (must be Hashable)
V = TypeVar('V')  # Represents the value type


def previous_element(dct: OrderedDict[K, V], key: K) -> V | None:
    """
    Get the previous element.

    Returns the value of the element that comes before the given key
    in an ordered dictionary.
    If the key is not in the dictionary, or if it is the first element
    in the dictionary, returns None.

    Arguments:
        dct: An ordered dictionary.
        key: The key of the element whose previous element we want to
        find.

    Returns:
        The value of the element that comes before the given key in
        the dictionary, or None if there is no such element.
    """
    if key in dct:
        key_list = list(dct.keys())
        idx = key_list.index(key)
        if idx == 0:
            result = None
        else:
            result = dct[key_list[idx - 1]]
    else:
        result = None
    return result



"""
UID Object.

Parent class for all objects that have a unique identifier (UID).
"""

from dataclasses import dataclass

from osmg.creators.uid import UIDGenerator


@dataclass
class UIDObject:
    """Base class for objects with a unique identifier (UID)."""

    uid_generator: UIDGenerator

    def __post_init__(self) -> None:
        """Post-initialization."""
        self._uid = self.uid_generator.new(self)

    @property
    def uid(self) -> int:
        """Get the UID."""
        return self._uid

    def __hash__(self) -> int:
        """Return the hash of the object based on its UID."""
        return hash(self.uid)

    def __eq__(self, other: object) -> bool:
        """
        Check equality based on the UID.

        Returns:
          True if it is equal, False otherwise.
        """
        if not isinstance(other, UIDObject):
            return False
        return self.uid == other.uid



"""Defines OpenSees Element interfrace objects."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

import numpy as np

from osmg.core.uid_object import UIDObject
from osmg.graphics.visibility import ElementVisibility

if TYPE_CHECKING:
    from osmg.core.common import numpy_array
    from osmg.mesh import Mesh
    from osmg.model_objects.node import Node
    from osmg.model_objects.section import ElasticSection, FiberSection
    from osmg.model_objects.uniaxial_material import UniaxialMaterial


@dataclass(repr=False)
class Element(UIDObject):
    """
    OpenSees element.

    https://openseespydoc.readthedocs.io/en/latest/src/element.html

    Attributes:
    ----------
        parent_component:
          the parent component assembly that this element belongs to.
        uid: the unique identifier of this element.
        nodes: the list of nodes that this element connects.

    """

    nodes: list[Node]

    def __post_init__(self) -> None:
        """Post-initialization."""
        self.visibility = ElementVisibility()
        super().__post_init__()


@dataclass(repr=False)
class ZeroLength(Element):
    """
    OpenSees ZeroLength element.

    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html

    """

    materials: list[UniaxialMaterial]
    directions: list[int]
    vecx: numpy_array | None
    vecyp: numpy_array | None
    enable_rayleigh: bool = field(default=True)

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        output = [
            'zeroLength',
            self.uid,
            *[n.uid for n in self.nodes],
            '-mat',
            *[m.uid for m in self.materials],
            '-dir',
            *self.directions,
        ]
        if self.enable_rayleigh:
            output.extend(
                [
                    '-doRayleigh',
                    1,
                ]
            )
        else:
            output.extend(
                [
                    '-doRayleigh',
                    0,
                ]
            )
        if self.vecx is not None or self.vecyp is not None:
            assert self.vecx is not None
            assert self.vecyp is not None
            output.extend(
                [
                    '-orient',
                    *self.vecx,
                    *self.vecyp,
                ]
            )

        return output

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'ZeroLength element object\n'
        res += f'uid: {self.uid}'
        res += 'Materials:'
        for mat, direction in zip(self.materials, self.directions):
            res += f'  {direction}: {mat.name}\n'
        res += f'vecx: {self.vecx}\n'
        res += f'vecyp: {self.vecyp}\n'
        return res


@dataclass(repr=False)
class TwoNodeLink(Element):
    """
    OpenSees TwoNodeLink element.

    https://openseespydoc.readthedocs.io/en/latest/src/twoNodeLink.html

    """

    materials: list[UniaxialMaterial]
    directions: list[int]
    vecyp: numpy_array | None = None

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        args = [
            'twoNodeLink',
            self.uid,
            *[n.uid for n in self.nodes],
            '-mat',
            *[m.uid for m in self.materials],
            '-dir',
            *self.directions,
        ]
        if self.vecyp is not None:
            args.extend(
                [
                    '-orient',
                    *self.vecyp,
                ]
            )
        return args

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'TwoNodeLink element object\n'
        res += f'uid: {self.uid}'
        res += 'Materials:'
        for mat, direction in zip(self.materials, self.directions):
            res += f'  {direction}: {mat.name}\n'
        if self.vecyp:
            res += f'vecyp: {self.vecyp}\n'
        return res


@dataclass(repr=False)
class Bar(Element):
    """
    Truss and Corotational Truss.

    Implements both of the following:
    OpenSees Truss Element
    https://openseespydoc.readthedocs.io/en/latest/src/trussEle.html
    OpenSees Corotational Truss Element
    https://openseespydoc.readthedocs.io/en/latest/src/corotTruss.html

    Params:
      transf_type: Linear or Corotational, determines what type of truss to use.
      area: Cross-sectional area.
      mat: Uniaxial material (for stress-strain relationship).
      outside_shape: Mesh defining the cross-section shape, used only for plots.
      weight_per_length: Weight per unit length, used when assigning self-weight.
      rho: Mass per unit length.
      cFlag: 0: lumped mass, 1: consistent mass.
      rFlag: 0: No Rayleigh damping, 1: Include Rayleigh damping.
    """

    transf_type: Literal['Linear', 'Corotational']
    area: float
    material: UniaxialMaterial
    outside_shape: Mesh | None = field(default=None)
    weight_per_length: float = field(default=0.00)
    rho: float = field(default=0.00)
    cflag: int = field(default=0)
    rflag: int = field(default=0)

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        elm_name = {'Linear': 'Truss', 'Corotational': 'corotTruss'}

        return [
            elm_name[self.transf_type],
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.area,
            self.material.uid,
            '-rho',
            self.rho,
            '-cMass',
            self.cflag,
            '-doRayleigh',
            self.rflag,
        ]

    def clear_length(self) -> float:
        """
        Clear length.

        Returns the clear length of the element (without the rigid
        offsets)

        Returns:
          The clear length.
        """
        p_i = np.array(self.nodes[0].coordinates)
        p_j = np.array(self.nodes[1].coordinates)
        return float(np.linalg.norm(p_i - p_j))

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        elm_name = {'Linear': 'Truss', 'Corotational': 'corotTruss'}
        res = ''
        res += f'{elm_name[self.transf_type]} element object\n'
        res += f'uid: {self.uid}\n'
        res += f'node_i.uid: {self.nodes[0].uid}\n'
        res += f'node_j.uid: {self.nodes[1].uid}\n'
        res += f'node_i.coordinates: {self.nodes[0].coordinates}\n'
        res += f'node_j.coordinates: {self.nodes[1].coordinates}\n'
        return res


@dataclass(repr=False)
class GeomTransf(UIDObject):
    """
    Geometric transformation.

    `y_axis` = None implies 2D space.

    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html
    https://openseespydoc.readthedocs.io/en/latest/src/geomTransf.html?highlight=geometric%20transformation

    """

    transf_type: Literal['Linear', 'Corotational', 'PDelta']
    offset_i: numpy_array
    offset_j: numpy_array
    x_axis: numpy_array
    y_axis: numpy_array | None
    z_axis: numpy_array

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        if self.y_axis is not None:
            # 3D case
            return [
                self.transf_type,
                self.uid,
                *self.z_axis,
                '-jntOffset',
                *self.offset_i,
                *self.offset_j,
            ]
        # 2D case
        return [
            self.transf_type,
            self.uid,
            '-jntOffset',
            *self.offset_i,
            *self.offset_j,
        ]


@dataclass(repr=False)
class ModifiedStiffnessParameterConfig:
    """Configuration parameters for ModifiedElasticBeam elements."""

    n_x: float | None
    n_y: float | None


@dataclass(repr=False)
class BeamColumnElement(Element):
    """Beamcolumn element."""

    section: ElasticSection
    geomtransf: GeomTransf

    def clear_length(self) -> float:
        """
        Clear length.

        Returns the clear length of the element (without the rigid
        offsets)

        Returns:
          The clear length.
        """
        p_i = np.array(self.nodes[0].coordinates) + self.geomtransf.offset_i
        p_j = np.array(self.nodes[1].coordinates) + self.geomtransf.offset_j
        return float(np.linalg.norm(p_i - p_j))


@dataclass(repr=False)
class ElasticBeamColumn(BeamColumnElement):
    """
    OpenSees Elastic Beam Column Element.

    https://openseespydoc.readthedocs.io/en/latest/src/elasticBeamColumn.html
    """

    modified_stiffness_config: ModifiedStiffnessParameterConfig | None = field(
        default=None
    )

    @staticmethod
    def _calculate_stiffness(n: float) -> tuple[float, float, float]:
        """
        Calculate stiffness parameters based on the given n value.

        Args:
            n: The stiffness parameter (n_x or n_y).

        Returns:
            A tuple containing (k11, k33, k44).
        """
        k44 = 6.0 * (1.0 + n) / (2.0 + 3.0 * n)
        k11 = (1.0 + 2.0 * n) * k44 / (1.0 + n)
        k33 = k11
        return k11, k33, k44

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
            The OpenSees arguments.
        """
        mod_params = self.modified_stiffness_config
        if mod_params:
            stiffness_x = (
                self._calculate_stiffness(mod_params.n_x)
                if mod_params.n_x is not None
                else None
            )
            stiffness_y = (
                self._calculate_stiffness(mod_params.n_y)
                if mod_params.n_y is not None
                else None
            )

            args = [
                'ModElasticBeam3d',
                self.uid,
                self.nodes[0].uid,
                self.nodes[1].uid,
                self.section.area,
                self.section.e_mod,
                self.section.g_mod,
                self.section.j_mod,
                self.section.i_y,
                self.section.i_x,
            ]

            # Append stiffness parameters based on availability
            if stiffness_x and not stiffness_y:
                args.extend([4.00, 4.00, 2.00, *stiffness_x, self.geomtransf.uid])
            elif stiffness_y and not stiffness_x:
                args.extend([*stiffness_y, 4.00, 4.00, 2.00, self.geomtransf.uid])
            elif stiffness_x and stiffness_y:
                args.extend([*stiffness_y, *stiffness_x, self.geomtransf.uid])

            return args

        # Default elastic beam column configuration if no modified stiffness
        return [
            'elasticBeamColumn',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.section.area,
            self.section.e_mod,
            self.section.g_mod,
            self.section.j_mod,
            self.section.i_y,
            self.section.i_x,
            self.geomtransf.uid,
        ]

    def ops_args_2d(self) -> list[object]:
        """
        Obtain the OpenSees arguments for a 2D model.

        Returns:
            The OpenSees arguments.
        """
        mod_params = self.modified_stiffness_config
        if mod_params:
            msg = 'Not implemented yet.'
            raise NotImplementedError(msg)

        # Default elastic beam column configuration if no modified stiffness
        return [
            'elasticBeamColumn',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.section.area,
            self.section.e_mod,
            self.section.i_x,
            self.geomtransf.uid,
        ]

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'elasticBeamColumn element object\n'
        res += f'uid: {self.uid}\n'
        res += f'node_i.uid: {self.nodes[0].uid}\n'
        res += f'node_j.uid: {self.nodes[1].uid}\n'
        res += f'node_i.coordinates: {self.nodes[0].coordinates}\n'
        res += f'node_j.coordinates: {self.nodes[1].coordinates}\n'
        res += f'offset_i: {self.geomtransf.offset_i}\n'
        res += f'offset_j: {self.geomtransf.offset_j}\n'
        res += f'x_axis: {self.geomtransf.x_axis}\n'
        res += f'y_axis: {self.geomtransf.y_axis}\n'
        res += f'z_axis: {self.geomtransf.z_axis}\n'
        res += f'section.name: {self.section.name}\n'
        return res


@dataclass
class BeamIntegration(UIDObject):
    """
    OpenSees beamIntegration parent class.

    https://openseespydoc.readthedocs.io/en/latest/src/beamIntegration.html?highlight=beamintegration

    """

    parent_section: ElasticSection | FiberSection = field(repr=False)


@dataclass
class Lobatto(BeamIntegration):
    """
    OpenSees Lobatto beam integration.

    https://openseespydoc.readthedocs.io/en/latest/src/Lobatto.html

    """

    n_p: int

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        return ['Lobatto', self.uid, self.parent_section.uid, self.n_p]


@dataclass
class DispBeamColumn(BeamColumnElement):
    """
    OpenSees dispBeamColumn element.

    https://openseespydoc.readthedocs.io/en/latest/src/ForceBeamColumn.html

    """

    integration: BeamIntegration

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        return [
            'dispBeamColumn',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.geomtransf.uid,
            self.integration.uid,
            # '-iter',  # can change it to forceBeamColumn here for testing
            # 50,
            # 1e-1
        ]

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'dispBeamColumn element object\n'
        res += f'uid: {self.uid}\n'
        res += f'node_i.uid: {self.nodes[0].uid}\n'
        res += f'node_j.uid: {self.nodes[1].uid}\n'
        res += f'node_i.coordinates: {self.nodes[0].coordinates}\n'
        res += f'node_j.coordinates: {self.nodes[1].coordinates}\n'
        res += f'offset_i: {self.geomtransf.offset_i}\n'
        res += f'offset_j: {self.geomtransf.offset_j}\n'
        res += f'x_axis: {self.geomtransf.x_axis}\n'
        res += f'y_axis: {self.geomtransf.y_axis}\n'
        res += f'z_axis: {self.geomtransf.z_axis}\n'
        res += f'section.name: {self.section.name}\n'
        return res


@dataclass(repr=False)
class LeadRubberX(Element):
    """
    OpenSees LeadRubberX element.

    https://openseespydoc.readthedocs.io/en/latest/src/LeadRubberX.html
    """

    f_y: float  # yield strength
    alpha: float  # post-yield stiffness ratio
    g_r: float  # shear modulus of bearing
    k_bulk: float  # bulk modulus of rubber
    d_1: float  # internal diameter
    d_2: float  # outer diameter
    t_s: float  # single steel shim layer thickness
    t_r: float  # single rubber layer thickness
    n_layers: int  # number of rubber layers
    x_1: float  # x component of local x-axis
    x_2: float  # y component of local x-axis
    x_3: float  # z component of local x-axis
    y_1: float  # x component of local y-axis
    y_2: float  # y component of local y-axis
    y_3: float  # z component of local y-axis
    k_c: float = field(default=10.0)  # cavitation parameter
    phi_m: float = field(default=0.5)  # damage parameter
    a_c: float = field(default=1.0)  # strength reduction parameter
    s_d_ratio: float = field(default=0.5)  # shear distance
    mass: float = field(default=0.00)  # element mass
    c_d: float = field(default=0.00)  # viscous damping coefficient
    t_c: float = field(default=0.00)  # cover thickness
    q_l: float = field(default=11200.00)  # density of lead (units: kg/m3)
    c_l: float = field(default=130.00)  # specific heat of lead (units: N-m/kg oC)
    k_s: float = field(
        default=50.00
    )  # thermal conductivity of steel (units: 50W/m oC)
    a_s: float = field(
        default=1.41e-5
    )  # thermal diffusivity of steel (units: 1.41e-5m2/s)
    tag_1: int = field(default=0)  # Include cavitation and post-cavitation
    tag_2: int = field(default=0)  # Include buckling load variation
    tag_3: int = field(default=0)  # Include horizontal stiffness variation
    tag_4: int = field(default=0)  # Include vertical stiffness variation
    tag_5: int = field(
        default=0
    )  # Include strength degradation in shear due to heating of lead core

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        return [
            'LeadRubberX',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.f_y,
            self.alpha,
            self.g_r,
            self.k_bulk,
            self.d_1,
            self.d_2,
            self.t_s,
            self.t_r,
            self.n_layers,
            self.x_1,
            self.x_2,
            self.x_3,
            self.y_1,
            self.y_2,
            self.y_3,
            self.k_c,
            self.phi_m,
            self.a_c,
            self.s_d_ratio,
            self.mass,
            self.c_d,
            self.t_c,
            self.q_l,
            self.c_l,
            self.k_s,
            self.a_s,
            self.tag_1,
            self.tag_2,
            self.tag_3,
            self.tag_4,
            self.tag_5,
        ]


@dataclass(repr=False)
class ElastomericX(Element):
    """
    OpenSees LeadRubberX element.

    https://openseespydoc.readthedocs.io/en/latest/src/LeadRubberX.html
    """

    f_y: float  # yield strength
    alpha: float  # post-yield stiffness ratio
    g_r: float  # shear modulus of bearing
    k_bulk: float  # bulk modulus of rubber
    d_1: float  # internal diameter
    d_2: float  # outer diameter
    t_s: float  # single steel shim layer thickness
    t_r: float  # single rubber layer thickness
    n_layers: int  # number of rubber layers
    x_1: float  # x component of local x-axis
    x_2: float  # y component of local x-axis
    x_3: float  # z component of local x-axis
    y_1: float  # x component of local y-axis
    y_2: float  # y component of local y-axis
    y_3: float  # z component of local y-axis
    k_c: float = field(default=10.0)  # cavitation parameter
    phi_m: float = field(default=0.5)  # damage parameter
    a_c: float = field(default=1.0)  # strength reduction parameter
    s_d_ratio: float = field(default=0.5)  # shear distance
    mass: float = field(default=0.00)  # element mass
    c_d: float = field(default=0.00)  # viscous damping coefficient
    t_c: float = field(default=0.00)  # cover thickness
    q_l: float = field(default=11200.00)  # density of lead (units: kg/m3)
    c_l: float = field(default=130.00)  # specific heat of lead (units: N-m/kg oC)
    k_s: float = field(
        default=50.00
    )  # thermal conductivity of steel (units: 50W/m oC)
    a_s: float = field(
        default=1.41e-5
    )  # thermal diffusivity of steel (units: 1.41e-5m2/s)
    tag_1: int = field(default=0)  # Include cavitation and post-cavitation
    tag_2: int = field(default=0)  # Include buckling load variation
    tag_3: int = field(default=0)  # Include horizontal stiffness variation
    tag_4: int = field(default=0)  # Include vertical stiffness variation
    tag_5: int = field(
        default=0
    )  # Include strength degradation in shear due to heating of lead core

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        return [
            'LeadRubberX',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.f_y,
            self.alpha,
            self.g_r,
            self.k_bulk,
            self.d_1,
            self.d_2,
            self.t_s,
            self.t_r,
            self.n_layers,
            self.x_1,
            self.x_2,
            self.x_3,
            self.y_1,
            self.y_2,
            self.y_3,
            self.k_c,
            self.phi_m,
            self.a_c,
            self.s_d_ratio,
            self.mass,
            self.c_d,
            self.t_c,
            self.q_l,
            self.c_l,
            self.k_s,
            self.a_s,
            self.tag_1,
            self.tag_2,
            self.tag_3,
            self.tag_4,
            self.tag_5,
        ]



"""Model Generator for OpenSees ~ visibility."""

from dataclasses import dataclass, field


@dataclass
class ElementVisibility:
    """
    Element visibility object.

    Controls whether an element is displayed in the plots
    and whether it is defined in OpenSees or not
    """

    hidden_when_extruded: bool = field(default=False)
    hidden_at_line_plots: bool = field(default=False)
    skip_opensees_definition: bool = field(default=False)
    hidden_basic_forces: bool = field(default=False)
